#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from . import K_Component    as KC
from . import C_colors       as CC
from . import M_DataAnalysis as MD
from   geopy.distance      import great_circle
from . import M_GetCountry as MG 
from . path_segmenter import path_segmenting

def pipeline_sequencing(lat_list,lon_list,max_length=10):
    '''

    '''
    lat_list2=[]
    lon_list2=[]

    for lat1,lat2,lon1,lon2 in zip(lat_list[:-1],lat_list[1:],lon_list[:-1],lon_list[1:]):
       length=great_circle((lat1,lon1),(lat2,lon2)).km
       if length>max_length:
           lon_avg,lat_avg=midpoint((lon1,lon2),(lat1,lat2))
           lat_tmp1,lon_tmp1=pipeline_sequencing([lat1,lat_avg],[lon1,lon_avg],max_length)
           lat_tmp2,lon_tmp2=pipeline_sequencing([lat_avg,lat2],[lon_avg,lon2],max_length)
           lat_list2.extend(lat_tmp1[:-1])
           lon_list2.extend(lon_tmp1[:-1])
           lat_list2.extend(lat_tmp2)
           lon_list2.extend(lon_tmp2)

       else:
           lat_list2.extend([lat1,lat2])
           lon_list2.extend([lon1,lon2])

       lat_list2=lat_list2[:-1]
       lon_list2=lon_list2[:-1]

    lat_list2.append(lat2)
    lon_list2.append(lon2)
    return lat_list2,lon_list2

def new_node_ID_nodes():
    for i in range(100000000):
        yield f'OSM_{i}'

def new_node_ID(x,y):
    for i in range(10000):
        yield x+'-'+y+'_'+str(i)

def old_node_IDs(id_list):
    for i in range(len(id_list)-1):
        yield id_list[i],id_list[i+1]

def create_segments_INET(pipelines,max_distance=10, verbose=True):
    "creates addition pipeline nodes and add them to the network"
    warning=False
    for pipeline in pipelines:
        if len(pipeline.lat) != len(pipeline.node_id):
            warning=True
    if warning:
        print(CC.Red+'Warning: length of id_nodes not equals length of coordinates per pipeline'+CC.End)
    countrydict=MG.CountryPolyDict()
 
    for pipe in pipelines:
        old_nodes = list(zip(pipe.lat,pipe.long))
        lat,long = pipeline_sequencing(pipe.lat,pipe.long,max_length=max_distance)
        new_nodes = list(zip(lat,long))
        old_ID_generator = old_node_IDs(pipe.node_id)
        new_node_ids = []
        first_id,second_id = next(old_ID_generator)
        new_ID_generator=new_node_ID(first_id,second_id)
        for i, new_node in enumerate(new_nodes[:-1]):
            if new_node in old_nodes:
                new_node_ids.append(first_id)
            else: new_node_ids.append(next(new_ID_generator))


            if new_nodes[i+1] in old_nodes:
                if i+1<(len(new_nodes)-1):
                    first_id,second_id = next(old_ID_generator)
                    new_ID_generator=new_node_ID(first_id,second_id)
        new_node_ids.append(second_id)
        pipe.node_id=new_node_ids
        pipe.lat=lat
        pipe.long=long
        old_countrycodes=pipe.country_code
        pipe.country_code=MG.GetCountry4List(long,lat,countrydict,old_countrycodes)
    if verbose==True:
        print('--Segmenting INET--')

    return pipelines


def create_segments(pipelines,max_distance=10):
    for pipe in pipelines:
        lat,long=pipeline_sequencing(pipe.lat,pipe.long,max_distance)
        pipe.lat=lat
        pipe.long=long
    return pipelines

def OSM_ShortSegmentRemove(Pipelines,maxlength):
    """
    Removes all pipelines with a length<maxlength[km]
    """
    for pipe in reversed(Pipelines):
#       if float(pipe.length) < maxlength:
       if float(pipe.param['length_km'])< maxlength:
           Pipelines.remove(pipe)
    pass

def OSM_Segmenting(line_long, line_lat, i, maxlength,verbose=False):
    """
    Divides a OSM Line into Segments with a length < maxlength [km] if possible
    """
    line_a_long = []
    line_b_long = []
    line_a_lat  = []
    line_b_lat  = []
    i = str(i)
    length = MD.routelength(line_long, line_lat)
    if verbose==True:
        print(f'Pipeline {i} -- Length:{length} km')

    if len(line_long) ==2:
        if float(length)>maxlength:
#            print(f'Pipeline {i} > {maxlength} km -> create interlines from interpol')
            line_lat,line_long=pipeline_sequencing(line_lat,line_long,5)
#            print(f'linelong {line_long}')
            line_long,line_lat=OSM_Segmenting(line_long,line_lat,i,maxlength)
            middle=int(round(len(line_long)//2))
            line_a_long, line_a_lat = OSM_Segmenting(line_long[0:middle],line_lat[0:middle],str(i)+'.1',maxlength)
            line_b_long, line_b_lat = OSM_Segmenting(line_long[middle:],line_lat[middle:],str(i)+'.2',maxlength)
            line_long = []
            line_lat  = []
            line_long.extend(line_a_long)
            line_long.extend(line_b_long)
            line_lat.extend(line_a_lat)
            line_lat.extend(line_b_lat)
            return line_long, line_lat

        else:
            xline_long = [line_long[0],line_long[-1]]
            xline_lat  = [line_lat[0],line_lat[-1]]
            return line_long, line_lat
    # Split lines (recursively) if they have more than 2 points and are longer then maxlength
    elif len(line_long) > 2:
        if float(length) > maxlength:
#            print(f'Pipeline {i} > {maxlength} km -> split this line')
            middle=int(round(len(line_long)//2))


            line_a_long, line_a_lat = OSM_Segmenting(
                                      line_long[0:middle+1],
                                      line_lat[0:middle+1],
                                      str(i)+'.1',maxlength)
            line_b_long, line_b_lat = OSM_Segmenting(
                                      line_long[middle:],
                                      line_lat[middle:],
                                      str(i)+'.2',maxlength)
            line_long = []
            line_lat  = []
            line_long.extend(line_a_long)
            line_long.extend(line_b_long)
            line_lat.extend(line_a_lat)
            line_lat.extend(line_b_lat)
            return line_long, line_lat
        else:
            xline_long = [line_long[0], line_long[-1]]
            xline_lat  = [line_lat[0], line_lat[-1]]
            return xline_long, xline_lat
    pass

def OSM_Pipe2Segment(pipe, i,length,verbose=False,compose=False):
    """
    Creates PipeSegments from a OSM-Pipeline
    """
    res_long, res_lat = OSM_Segmenting(pipe.long,pipe.lat,i,length,verbose=verbose)
    if verbose==True:
        print(f'res_lat: {res_lat}')
        print(f'res_long: {res_long}')

    # (int i = 0; i < list.length(); i+=2)
    pipesegments = []


    path_coords=list(zip(pipe.param['path_lat'],
                         pipe.param['path_long']))
    
    for i in range(0,len(res_long)-1,2):

        id   = pipe.id+'_Seg_'+str(int(i/2+1))
        lat  = [res_lat[i],res_lat[i+1]]
        long = [res_long[i],res_long[i+1]]
        name = pipe.id+'_Seg_'+str(int(i/2+1))

        source_id = pipe.source_id
        # try:
        #     node_id  = [pipe.node_id[i],pipe.node_id[i+1]]
        # except:
        #     node_id = 'no IDs created by pipelinesequencing'
        

        # if first_node not in current_nodes[0]
        # if current_nodes==[]:    

           # node_id=[next(new_id_generator),next(new_id_generator)]
        node_id=[]
        pipe_coords=list(zip(lat,long))
        #print('pipe_coords: ',pipe_coords)
        #print('path_coords: ',path_coords)
        if type(pipe.country_code)==str:
            pipe.country_code=[pipe.country_code]
#          self, id, name, source_id, node_id, lat, long, country_code, tags,**param
        new_path_coords=path_segmenting(pipe_coords,path_coords)
        #print('new_path:',new_path_coords)
        
        path_lat=[coord[0] for coord in new_path_coords]
        #print('pathlat:',path_lat)
        path_long=[coord[1] for coord in new_path_coords]
        
        new_param=dict(pipe.param)
        new_param['path_lat']=path_lat
        new_param['path_long']=path_long  
        country_code     = list(set(pipe.country_code))
        tags             = pipe.tags
        pipesegment      = KC.OSMComponent(id,
                                           name,
                                           source_id,
                                           node_id,
                                           lat,
                                           long,
                                           country_code,
                                           tags,
                                           param = new_param,
                                           uncertainty=pipe.uncertainty,
                                           method=pipe.method)
#        pipesegment.lat  = lat
#        pipesegment.long = long
        pipesegments.append(pipesegment)

    return pipesegments

def OSM_Pipe2Pipe(pipe, i,length,verbose=False):
    """
    Creates PipeSegments from a OSM-Pipeline
    """
    res_long, res_lat = OSM_Segmenting(pipe.long,pipe.lat,i,length,verbose=verbose)
    if verbose==True:
        print(f'res_lat: {res_lat}')
        print(f'res_long: {res_long}')

    # (int i = 0; i < list.length(); i+=2)

    
    new_long=[]
    for i in range(0,len(res_long)-1,2):
        a,b = res_long[i:i+2]
        save=b
        new_long.append(a)
    new_long.append(save)

    new_lat=[]
    for i in range(0,len(res_long)-1,2):
        a,b = res_lat[i:i+2]
        save=b
        new_lat.append(a)
    new_lat.append(save)

    id   = pipe.id
    lat  = new_lat
    long = new_long
    name = pipe.id

    source_id = pipe.source_id
        # try:
        #     node_id  = [pipe.node_id[i],pipe.node_id[i+1]]
        # except:
        #     node_id = 'no IDs created by pipelinesequencing'
        

        # if first_node not in current_nodes[0]
        # if current_nodes==[]:    

           # node_id=[next(new_id_generator),next(new_id_generator)]
    node_id=[]

    if type(pipe.country_code)==str:
        pipe.country_code=[pipe.country_code]
        
#          self, id, name, source_id, node_id, lat, long, country_code, tags,**param


    country_code     = list(set(pipe.country_code))
    tags             = pipe.tags
    pipeline     = KC.OSMComponent(id,
                                       name,
                                       source_id,
                                       node_id,
                                       lat,
                                       long,
                                       country_code,
                                       tags,
                                       param = pipe.param,
                                       uncertainty=pipe.uncertainty,
                                       method=pipe.method)
#        pipesegment.lat  = lat
#        pipesegment.long = long
    # pipelines.append(pipeline)

    return pipeline


def OSM_Pipelines2Segments(Component,length,verbose=False):
    """
    Creates PipeSegments from a list of OSM-Pipelines
    """
    linelist=[]
               
        
                 
    for i,pipe in enumerate(Component):
        lines=OSM_Pipe2Segment(pipe,i,length,verbose=verbose)
#        print(lines)
        for line in lines:
            linelist.append(line)

    # Component.clear()
    # Component.extend(linelist)
    return linelist

def OSM_Pipelines2Pipelines(Component,length,verbose=False):
    """
    Creates PipeSegments from a list of OSM-Pipelines
    """
    linelist=[]
               
    for i,pipe in enumerate(Component):
        line=OSM_Pipe2Pipe(pipe,i,length,verbose=verbose)
        linelist.append(line)
#        print(lines)


    # Component.clear()
    # Component.extend(lines)
    return linelist


def OSM_PipelineSegmenting(Component,length=40,minlength=3.0,verbose=False,composed=False):
    """
    Segmenting of OSM-PipeLines to PipeSegments with a length < maxlength
    Removing Segments < minlength
    length in deg^2
    """
    if verbose==True:
        print(CC.Cyan + 'Start Pipesegmenting:'+ CC.End)
    if composed==False:
        NewComponent=OSM_Pipelines2Segments(Component,length,verbose=verbose)
        if verbose==True:
            print(CC.Cyan + 'Recalculate lengths for all Segments' + CC.End)
        MD.set_lengths(NewComponent)
        if verbose==True:
            print(CC.Cyan + 'Remove all short Segments '+ str(minlength) + CC.End)
        OSM_ShortSegmentRemove(NewComponent,minlength)
    else:
        NewComponent=OSM_Pipelines2Pipelines(Component,length,verbose=verbose)
        


    return NewComponent


def average(a,b):
    res_a = (a[0]+a[1])/2
    res_b = (b[0]+b[1])/2
    return res_a,res_b


def midpoint(a,b):
    from math import pi,cos,sin,atan2,sqrt

    lon1 = a[0]/180*pi
    lon2 = a[1]/180*pi
    lat1 = b[0]/180*pi
    lat2 = b[1]/180*pi

    Bx     = cos(lat2) * cos(lon2-lon1)
    By     = cos(lat2) * sin(lon2-lon1)
    latMid = atan2(sin(lat1) + sin(lat2), sqrt( (cos(lat1)+Bx)*(cos(lat1)+Bx) + By*By ) )
    lonMid = lon1 + atan2(By, cos(lat1) + Bx)
    return round(lonMid/pi*180,5),round(latMid/pi*180,5)

#%%
if __name__=="__main__":
    import os
    os.chdir('..')

    pipes=create_segments_INET(INET.PipeSegments)


    # a=[40,53.939785, 54.801878, 54.80609]
    # b=[10,9.784432, 9.289041, 9.29094]

    # print(pipeline_sequencing(b,a,200))
    # print('ada')

#     lat: [56.962682700000244, 56.96265560000024, 56.96261540000024, 56.962586600000236]
# long: [24.03559059999993, 24.03558749999993, 24.03675789999993, 24.03675569999993]



from quantulum3 import parser

def extract_values(tag_dict_list, parameter,verbose=False):

    counterlist = []
    valuelist = []
    unitylist = []

    for i, tag_dict in enumerate(tag_dict_list):
        if parameter in tag_dict:
            if parameter in ['operator','start_operator']:
                value=tag_dict[parameter]
                valuelist.append(value)
            else:    
                value = parser.parse(tag_dict[parameter])
                print(tag_dict[parameter])
                if value != []:
                    if verbose==True:
                        print(i, tag_dict[parameter], value[0].unit.name)
                    unitylist.append(value[0].unit.uri)
                    valuelist.append(value[0].value)
            if value != []:
                counterlist.append(i)
               
                
    return counterlist, valuelist, unitylist

def values_to_pipeline(
    elements, pipe_numbers, values, value_name):
    for i, value in zip(pipe_numbers, 
                           values):
        elements[i].param.update({value_name: value})
        elements[i].uncertainty.update({value_name: 0})
        elements[i].method.update({value_name: "OSM"})
    pass

def values_to_pipeline_diameters(
    elements, pipe_numbers, diameter_values, diameter_unities
):
   

    for i, diameter, unit in zip(pipe_numbers, 
                                 diameter_values, 
                                 diameter_unities):

        if unit == "Centimeter":
            diameter = diameter * 10
        if unit == "Meter":
            diameter = diameter * 100
        if unit == "Inch":
            diameter = diameter * 25.4
        if unit == "Dimensionless_quantity" and diameter < 2:
            diameter = diameter * 1000

        elements[i].param.update({"diameter_mm": int(diameter)})
        elements[i].uncertainty.update({"diameter_mm": 0})
        elements[i].method.update({"diameter_mm": "OSM"})
    pass

def values_to_pipeline_pressures(
    elements, pipe_numbers, pressure_values, pressure_unities):
    for i, pressure, unit in zip(pipe_numbers, 
                                 pressure_values, 
                                 pressure_unities):

        if unit == "MPa":
            pressure = pressure * 10

        if pressure > 1000:
            pressure = None
                                             
        elements[i].param.update({"max_pressure_bar": pressure})
        elements[i].uncertainty.update({"max_pressure_bar": 0})
        elements[i].method.update({"max_pressure_bar": "OSM"})
    pass
        
def extract_tags(NET,verbose=False):
    tags_pipelines = NET.all("PipeLines", "tags", verbose=False)
    
    tags_pipesegments = NET.all("PipeSegments", "tags", verbose=False)

    component_names=['PipeSegments','PipeLines']
    components=[NET.PipeSegments,NET.PipeLines]
    tags_components=[tags_pipesegments,tags_pipelines]
  
    for tags_component,component,component_name in zip(tags_components,components,component_names):
    # Export Values to PipeSegments
        diameter_lines, diameter_values, diameter_unities = extract_values(
            tags_component, "diameter")
            
        values_to_pipeline_diameters(
            component, diameter_lines, diameter_values, diameter_unities)
            
        pressure_lines, pressure_values, pressure_unities = extract_values(
            tags_component, "pressure")
            
        values_to_pipeline_pressures(
            component, pressure_lines, pressure_values, pressure_unities)

        operator_lines, operator_values, operater_unities = extract_values(
            tags_component, "operator")

        values_to_pipeline(
            component, operator_lines, operator_values,"operator")

        operator_lines, operator_values, operater_unities = extract_values(
            tags_component, "start_date")

        values_to_pipeline(
            component, operator_lines, operator_values,"start_date")

            
        num_pipe = len(component)
        if verbose==True:    

            print(len(diameter_lines), " diameter values for ", 
                  num_pipe, component_name)
            print(len(pressure_lines), " pressure values for ", 
                  num_pipe, component_name)
    # Export Values to PipeSegments
    return NET



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
M_GetCountry
------------

Returns country codes from GeoCoordinates
"""

import shapefile
import sys
from pathlib  import Path
from shapely.geometry import Point, Polygon
import multiprocessing as mp

#Context Manager for time measurement
import time
import os
import contextlib


#Example for Use:
#    countrypolydict=CountryPolyDict()
#    for i,pipeline in enumerate(OSM.PipeLines):
#        countrys=CountryCheckfromList(pipeline.long,pipeline.lat,countrypolydict)
#        print(i,countrys)


def get_CC_List():
    countrylist = ['BA', 'DE', 'ES', 'FR', 'GB', 'IT', 'NL', 'AT', 'BE', 'CZ', 'DK',
                   'FI', 'GR', 'HU', 'LV', 'LI', 'LT', 'LU', 'NO', 'PL', 'CH', 'PT',
                   'RO', 'IE', 'HR', 'SI', 'SK', 'RS', 'ME', 'MK', 'BG', 'AL', 'CY',
                   'MT', 'SE', 'IS', 'EE', 'TR', 'BY', 'MD', 'AM', 'AZ', 'GE', 'SA',
                   'UA', 'RU', 'XX']
    return countrylist


def get_CC_List_EU27():
    countrylist = [ 'BE', 'BG', 'DK', 'DE', 'EE', 'FI', 'FR', 'GR', 'IE', 'IT',
                   'HR', 'LV', 'LT', 'LU', 'NL', 'AT', 'PL', 'PT', 'RO', 'SK', 'ES',
                   'SE', 'CZ', 'HU', 'CY', 'XX']

         # 'GB',
         #           'LI',  'NO', 'CH',
         #            'SI', 'RS', 'ME', 'MK', 'AL',
         #           'MT', 'IS', 'TR', 'BY', 'MD', 'AM', 'AZ', 'GE', 'SA',
         #           'UA', 'RU']
    return countrylist

def get_CC_List_EU28():
    countrylist = [ 'BE', 'BG', 'DK', 'DE', 'EE', 'FI', 'FR', 'GR', 'IE', 'IT',
                   'HR', 'LV', 'LT', 'LU', 'NL', 'AT', 'PL', 'PT', 'RO', 'SK', 'ES',
                   'SE', 'CZ', 'HU', 'CY', 'XX', 'GB']

         #            'SI', 'RS', 'ME', 'MK', 'AL', 'LI',  'NO', 'CH', 'UA', 'RU'
         #           'MT', 'IS', 'TR', 'BY', 'MD', 'AM', 'AZ', 'GE', 'SA',]
    return countrylist


def get_CC_List_EU28_exXX():
    countrylist = [ 'BE', 'BG', 'DK', 'DE', 'EE', 'FI', 'FR', 'GR', 'IE', 'IT',
                   'HR', 'LV', 'LT', 'LU', 'NL', 'AT', 'PL', 'PT', 'RO', 'SK', 'ES',
                   'SE', 'CZ', 'HU', 'CY', 'GB']
         #            'SI', 'RS', 'ME', 'MK', 'AL', 'LI',  'NO', 'CH', 'UA', 'RU'
         #           'MT', 'IS', 'TR', 'BY', 'MD', 'AM', 'AZ', 'GE', 'SA',]

    return countrylist



def get_CC_List_EU():
    countrylist = [ 'BE', 'BG', 'DK', 'DE', 'EE', 'FI', 'FR', 'GR', 'IE', 'IT',
                   'HR', 'LV', 'LT', 'LU', 'NL', 'AT', 'PL', 'PT', 'RO', 'SK', 'ES',
                   'SE', 'CZ', 'HU', 'CY', 'GB', 'CH', 'NO',  'XX', 'AL', 'LI', 'SI',
                   'RS', 'ME','MK', 'MT', 'IS', 'TR',]

                    #   'BY', 'MD', 'AM', 'AZ', 'GE', 'SA',
                    # 'UA', 'RU',]
    return countrylist




def get_CC_List_XX():
    countrylist = ['BA', 'DE', 'ES', 'FR', 'GB', 'IT', 'NL', 'AT', 'BE', 'CZ',
                   'DK', 'FI', 'GR', 'HU', 'LV', 'LI', 'LT', 'LU', 'NO', 'PL',
                   'CH', 'PT', 'RO', 'IE', 'HR', 'SI', 'SK', 'RS', 'ME', 'MK',
                   'BG', 'AL', 'CY', 'MT', 'SE', 'IS', 'EE', 'TR', 'BY', 'MD',
                   'AM', 'AZ', 'GE', 'SA', 'RU', 'UA', 'DZ', 'TN', 'KZ', 'MA',
                   'EG', 'JO', 'SY', 'IQ', 'IR', 'KW', 'IL', 'LY',]
    return countrylist



@contextlib.contextmanager
def benchmark(name):
    start = time.time()
    yield
    print('{} {:.2f}s'.format(name, time.time() - start))




def CountryPolyDict(FileName_Map,predicted_countrycodes = ''):
    '''Creates and Returns Dictionary {countrycode:list_of_polygons_for_that_country}
    Uses Shapefile
    Only Countries in the list country tested'''

    countrypolydict={}
#    if sys.platform == 'win32':
#       RelDirName      = './TM_World_Borders'
#        dataFolder      = Path.cwd()
#        filename        = dataFolder /  RelDirName
#        FileName_Map    = str(filename / 'TM_WORLD_BORDERS-0.3.shp')
#    else:
#        FileName_Map     = os.path.join(os.getcwd(),'TM_World_Borders/TM_WORLD_BORDERS-0.3.shp')


    #print(os.path.isfile(FileName_Map))

    sf          = shapefile.Reader(FileName_Map, encoding="latin1")
    polygon     = sf.shapes()
    countrylist = get_CC_List()

    #speed up the routine if countrycode is correctly predicted
    countrys=[]
    if predicted_countrycodes!='':
        if isinstance(predicted_countrycodes,str):
            predicted_countrycodes=[predicted_countrycodes]
            for predicted_countrycode in predicted_countrycodes:
                if predicted_countrycode in countrylist:
                    countrys.append(predicted_countrycode)
                    countrylist.remove(predicted_countrycode)
            countrys.extend(countrylist)
    else:
        countrys=countrylist


    for i,poly in enumerate(polygon):
            if sf.record(i)[1] in countrys:
                geoshapelist=[]
                for shape in poly.__geo_interface__['coordinates']:
                    if isinstance(shape, list):
                        geoshape=Polygon(shape[0])
                    else:
                        geoshape=Polygon(shape)
                    geoshapelist.append(geoshape)
                countrypolydict.update({sf.record(i)[1]:geoshapelist})
    return countrypolydict




def GetCountry(TM_World_Borders_file,long, lat, countrypolydict = '', predicted_countrycodes = ''):
    '''Get and return countrycode of the point(long,lat) else returns '??' as countrycode'''

    # get default country codes for europe
    if countrypolydict == '':
        countrypolydict = CountryPolyDict(TM_World_Borders_file,predicted_countrycodes)

    p1          = Point(long,lat)
    res_country = 'XX'
    found       = False

    for country in countrypolydict:
        for shape in reversed(range(len(countrypolydict[country]))):
            if p1.within(countrypolydict[country][shape]):
                res_country = country
                found       = True
                break
        if found:
            break
    return res_country




def GetCountries(long, lat, countrypolydict = '', predicted_countrycodes = ''):
    '''Get and return countrycode of the point(long,lat) else returns '??' as countrycode'''

    # get default country codes for europe
    if countrypolydict == '':
        countrypolydict = CountryPolyDict(TM_World_Borders_file,predicted_countrycodes)

    res_country = []
    N_long      = len(long)

    for idx in range(N_long):
        res_country.append('XX')

    for idx in range(N_long):
        p1          = Point(long[idx], lat[idx])
        found       = False

        for country in countrypolydict:
            for shape in reversed(range(len(countrypolydict[country]))):
                if p1.within(countrypolydict[country][shape]):
                    res_country[idx]    = country
                    found               = True
                    break
            if found:
                break


    return res_country




def GetCountry4List(TM_World_Borders_file,long_list, lat_list, countrypolydict, predicted_countrycodes = ''):
    '''Calls CountryCheck for list of Points (e.g Pipelinepoints)
    Gets the countrypolydict from CountryPolyDict()
    Returns countrycodeslist'''

    countrycodelist=[]
    for long,lat in zip(long_list,lat_list):
        country=GetCountry(TM_World_Borders_file,long,lat,countrypolydict,predicted_countrycodes)
        countrycodelist.append(country)
    return countrycodelist




def GetCountry4Component(component, countrypolydict='',predicted_countrycodes=''):
    '''Get countrycodes 4 all geopoints of a Netclass component e.g. OSM.PipeLines
    Returns list of list with country_codes'''

    if countrypolydict=='':
        countrypolydict=CountryPolyDict()
    with mp.Pool() as pool:
        jobs = [pool.apply_async(GetCountry4List, args=(element.long, element.lat,
                countrypolydict,predicted_countrycodes)) for element in component]
        results=[x.get() for x in jobs]

    return results




#Quicktest of function
if __name__=="__main__":
    import os
    os.chdir('..')
    dicta=CountryPolyDict()
    print(GetCountry(-6.28,53.35,dicta,['IE','ES']))  #IE
    print(GetCountry(-3.2100984,40.3900748,dicta)) #ES




'''
osmscigrid
===========

Filter for OpenStreetMap Protobuf data (aka `.pbf` files).

For convenience, the toplevel module :mod:`osm2scigrid_gas` links to the most
relevant classes and functions of this library:

.. autosummary::
    :nosignatures:
'''

import configparser
import os
from configparser import ExtendedInterpolation
from pathlib import Path


import pkg_resources
import os

resource_package = __name__
resource_path = '/'.join(('Setup', 'Setup_OSM.ini'))  # Do not use os.path.join()
template = pkg_resources.resource_string(resource_package, resource_path)

if not os.path.exists(os.path.dirname('Setup/')):
    os.makedirs(os.path.dirname('Setup/'))
    open(resource_path,"wb+").write(template)

if not os.path.exists(os.path.dirname('Data/')):
    os.makedirs(os.path.dirname('Data/'))

if not os.path.exists(os.path.dirname('../../TM_World_Borders/')):
    os.makedirs(os.path.dirname('../../TM_World_Borders/'))




from . import K_Netze, K_Component 
from . PosFind import find_pos_StringInList
import os
import pandas as pd
import numpy as np
from pathlib import Path
import math

def read(RelDirName = 'Eingabe/CSV/', NumDataSets = 1e+100, 
          skiprows = [], NetzName = None, NameStart = 'Gas'):
    """Description:
    ------------
        Reads Data from folder CSV_Path into Grid
        Grid = Instance of Netz Class

    Input Parameter:
    ----------------
        RelDirName    string containing path name of data location [default: 'Eingabe/CSV/']
        NumDataSets   Number of elements to be read for each component [default: 1e+100]
        skiprows      number of rows to skip [default: []]
        NetzName      String containing the nme of the network
                      [Default = None]
        NameStart     String, containing the file name start of the files
                      [Default = 'Gas']

        SourceName
    Return Parameters:
    ------------------
        Grid            instance of class K_Netze.Netz, populated with
                         data from CSV files  """
    # Dir name stuff
    DirName     = Path.cwd() /  RelDirName
    if not os.path.isdir(DirName):
        print("RelDirName does not exist")

    Grid        = K_Netze.NetComp()
    FileList    = K_Netze.NetComp().CompLabels()
    for key in FileList:
        count       = 0
        filename    = NameStart + '_' + key + '.csv'
        CSV_File    = str(DirName / filename)

        # Z set to zero if file does not exist
        Z           = CSV_2_list(CSV_File, skiprows = skiprows)
        if len(Z) > 0:
            for entry in Z:
                Keys    = list(entry.keys())
                Vals    = list(entry.values())
                for ii in range(len(Vals)):
                    if Vals[ii] == 'None':
                        Vals[ii] = None
                    elif type(Vals[ii]) is float:
                        if math.isnan(Vals[ii]):
                            Vals[ii] = None
                    else:
                        try:
                            Vals[ii] = float(Vals[ii])
                        except:
                            pass

                pos_Id   = find_pos_StringInList('id', Keys)
                pos_Name = find_pos_StringInList('name', Keys)
                pos_SId  = find_pos_StringInList('source_id', Keys)
                pos_Node = find_pos_StringInList('node_id', Keys)
                pos_CC   = find_pos_StringInList('country_code', Keys)
                pos_lat  = find_pos_StringInList('lat', Keys)
                pos_long = find_pos_StringInList('long', Keys)
                pos_comm = find_pos_StringInList('comment', Keys)
                pos_para = find_pos_StringInList('param', Keys)
                pos_meth = find_pos_StringInList('method', Keys)
                pos_unce = find_pos_StringInList('uncertainty', Keys)
                pos_tags = find_pos_StringInList('tags', Keys)

                # in case that license information is in file
                pos_license = find_pos_StringInList('license', Keys)

                del entry['id']
                try:
                    del entry['name']
                
                    del entry['source_id']
                    del entry['node_id']
                    del entry['country_code']

                    del entry['lat']
                    del entry['long']
                except: 
                    #print('warning no value to delete')
                    pass
                try:
                    del entry['comment']
                    del entry['param']
                    del entry['method']
                except:
                    #print('warning no  comment to delete')
                    pass
                try:
                    del entry['uncertainty']
                    del entry['tags']

                except:
                    #print('warning no  uncertainty to delete')
                    pass
                # in case that license information is in file
                try:
                    del entry['license']
                except:
                    pass


                name        = Vals[pos_Name[0]]
                source_id   = makeList(Vals[pos_SId[0]])
                node_id     = makeList(Vals[pos_Node[0]])
                country_code = makeList(Vals[pos_CC[0]])
                
                try:
                    id          = Vals[pos_Id[0]]

                except:
                    id = id

                lat         = Vals[pos_lat[0]]
                if isinstance(lat, str):
                    lat         = eval(lat)

                long        = Vals[pos_long[0]]
                if isinstance(long, str):
                    long        = eval(long)

                try:
                    comment     = Vals[pos_comm[0]]
                except:
                    #print('warning no  comment to read')
                    comment={}
                    pass
                try:
                    param       = eval(Vals[pos_para[0]].replace(': nan,', ': float(\'nan\'),'))
                    method      = eval(Vals[pos_meth[0]].replace(': nan,', ': float(\'nan\'),'))
                except:
                    #print('warning no  value to delete')
                    pass
                try:
                    uncertainty = eval(Vals[pos_unce[0]].replace(': nan,', ': float(\'nan\'),'))
                    tags        = eval(Vals[pos_tags[0]].replace(': nan,', ': float(\'nan\'),'))

                except:
                    #print('warning no  uncertainty to delete')
                    uncertainty={}
                    pass
                # in case that license information is in file
                if len(pos_license) > 0:
                    license_val = eval(Vals[pos_license[0]].replace(': nan,', ': float(\'nan\'),'))
                else:
                    license_val = {}

                Grid.__dict__[key].append(K_Component.__dict__[key](id = id,
                             name       = name,
                             source_id  = source_id,
                             node_id    = node_id,
                             country_code = country_code,
                             param      = param,
                             lat        = lat,
                             long       = long,
                             method     = method,
                             uncertainty = uncertainty,
                             tags       = tags,
                             comment    = comment))
                # In case that license information is given
                if license_val != {}:
                    Grid.__dict__[key][count].license = license_val
                count = count + 1
                if count >= NumDataSets:
                    break
        else:
            Grid.__dict__[key]      = []
    Grid.SourceName = [NetzName]
    return Grid

def CSV_2_list(file, skiprows = []):
    """Description:
    ------------
        Liest ein Netzattribute aus einer CSV-datei (file) als Dataframe ein
        und liefert eine Liste zurück
    Parameter:
    -----------
        file = filepath+name.csv (des Netzattributes)
    Needs:
    -------
        Panda Library
    called by:
    -----------
        read_CSV()
    """
    if os.path.isfile(file):
        with open(file,'rt', encoding = 'iso-8859-15', errors = 'ignore') as csvfile:
            df = pd.read_csv(csvfile, sep =';', skiprows = skiprows)
            liste = []
            liste = df.T.to_dict().values()

            return liste
    else:
        return []

def makeList(stringVal):
    """Function to convert a string values into a list"""

    listVal = []
    if stringVal != None:
        if isinstance(stringVal, str):
            if '[\'' in stringVal and '\']' in stringVal:
                stringVal = stringVal[1:-2]
                stringVal   = stringVal
                stringVal   = stringVal.replace('\', \'', '\',\'')
                for val in stringVal.split(','):
                    if val[0] == '\'':
                        val = val[1:]
                    if val[-1] == '\'':
                        val = val[:-1]
                    listVal.append(val)
            else:
                listVal = stringVal
        else:
            listVal = stringVal

    else:
        listVal = stringVal

    return listVal



from   osmscigrid.M_GetCountry import CountryPolyDict, GetCountry4List
from geopy.distance import great_circle
from math import pi,cos,sin,atan2,sqrt

TM_World_Borders_file='../TM_World_Borders/TM_WORLD_BORDERS-0.3.shp'

def Segments2Lines(Netz):
    Segments=Netz.__dict__['PipeSegments']
    for Segment in Segments:
        if len(Segment.lat)!=len(Segment.node_id):
           startid=Segment.node_id[0]
           endid=Segment.node_id[1]
           
           new_node_id=[]
           new_node_id.append(startid)
           for i in range(len(Segment.lat)-2):
               new_node_id.append(startid+'-'+str(i+1)+'-'+endid)
           new_node_id.append(endid)
           Segment.node_id=new_node_id
    return Netz

def remove_double_nodes(Netz):
    Segments=Netz.__dict__['PipeSegments']
    for Segment in Segments:
        Segment.lat=Segment.lat[0::2].append(Segment.lat[-1])
        Segment.long=Segment.long[0::2].append(Segment.long[-1])
    return Netz

def rename_nodes(Netz,prestring='LKD_'):
    newnodelist=[]
    complist=Netz.CompLabels()
    for component in complist:
        for element in Netz.__dict__[component]:
            newnodelist=[]
            for node in element.node_id:
                newnodelist.append(prestring+node)
            element.node_id=newnodelist
            
    complist=Netz.CompLabelsSpot()+['Nodes',]
    for component in complist:
        for element in Netz.__dict__[component]:
            element.id=(prestring+element.id)
            
    pass


def pipeline_sequencing(lat_list,lon_list,max_length=10):
    '''

    '''
    lat_list2=[]
    lon_list2=[]

    for lat1,lat2,lon1,lon2 in zip(lat_list[:-1],lat_list[1:],lon_list[:-1],lon_list[1:]):
       length=great_circle((lat1,lon1),(lat2,lon2)).km
       if length>max_length:
           lon_avg,lat_avg=midpoint((lon1,lon2),(lat1,lat2))
           lat_tmp1,lon_tmp1=pipeline_sequencing([lat1,lat_avg],[lon1,lon_avg],max_length)
           lat_tmp2,lon_tmp2=pipeline_sequencing([lat_avg,lat2],[lon_avg,lon2],max_length)
           lat_list2.extend(lat_tmp1[:-1])
           lon_list2.extend(lon_tmp1[:-1])
           lat_list2.extend(lat_tmp2)
           lon_list2.extend(lon_tmp2)

       else:
           lat_list2.extend([lat1,lat2])
           lon_list2.extend([lon1,lon2])

       lat_list2=lat_list2[:-1]
       lon_list2=lon_list2[:-1]

    lat_list2.append(lat2)
    lon_list2.append(lon2)
    return lat_list2,lon_list2


def new_node_IDs(Name):
    for i in range(1000000000):
        yield f'{Name}_{i}'


def new_node_ID_nodes():
    for i in range(100000000):
        yield f'SEQ_{i}'

def new_node_ID(x,y):
    for i in range(10000):
        yield x+'-'+y+'_'+str(i)

def old_node_IDs(id_list):
    for i in range(len(id_list)-1):
        yield id_list[i],id_list[i+1]
def average(a,b):
    res_a = (a[0]+a[1])/2
    res_b = (b[0]+b[1])/2
    return res_a,res_b


def midpoint(a,b):

    lon1 = a[0]/180*pi
    lon2 = a[1]/180*pi
    lat1 = b[0]/180*pi
    lat2 = b[1]/180*pi

    Bx     = cos(lat2) * cos(lon2-lon1)
    By     = cos(lat2) * sin(lon2-lon1)
    latMid = atan2(sin(lat1) + sin(lat2), sqrt( (cos(lat1)+Bx)*(cos(lat1)+Bx) + By*By ) )
    lonMid = lon1 + atan2(By, cos(lat1) + Bx)
    return round(lonMid/pi*180,5),round(latMid/pi*180,5)


def paths2nodes(pipelines,IDstring='New',max_distance=10, verbose=True):
    "keep first and last node as the are the only real nodes"
    "create intermediate nodes"
    warning=False
    for pipeline in pipelines:
        if len(pipeline.lat) != len(pipeline.node_id):
            warning=True
    if warning:
        print('Warning: length of id_nodes not equals length of coordinates per pipeline')
    countrydict=CountryPolyDict(TM_World_Borders_file)
    pipenodes=[]
    new_ID_generator=new_node_IDs(IDstring)
    for pipe in pipelines:
        old_nodes = list(zip(pipe.lat,pipe.long))
        lat,long = pipeline_sequencing(pipe.lat,pipe.long,max_length=max_distance)
        new_nodes = list(zip(lat,long))
        #old_ID_generator = old_node_IDs(pipe.node_id)
        new_node_ids = []
        first_id,last_id = pipe.node_id[0],pipe.node_id[-1]
        new_node_ids.append(first_id)
        for i, new_node in enumerate(new_nodes[1:-1]):
            #if new_node in old_nodes:
            #    new_node_ids.append(first_id)
            new_node_ids.append(next(new_ID_generator))


            # if new_nodes[i+1] in old_nodes:
                # if i+1<(len(new_nodes)-1):
                    # first_id,second_id = next(old_ID_generator)
                    # new_ID_generator=new_node_ID(first_id,second_id)
        new_node_ids.append(last_id)
        pipe.node_id=new_node_ids
        pipe.lat=lat
        pipe.long=long
        old_countrycodes=pipe.country_code
        pipe.country_code=GetCountry4List(TM_World_Borders_file,long,lat,countrydict,old_countrycodes)
    if verbose==True:
        print('--Segmenting INET--')

    return pipelines




def create_segments_INET(pipelines,max_distance=10, verbose=True):
    "creates addition pipeline nodes and add them to the network"
    warning=False
    for pipeline in pipelines:
        if len(pipeline.lat) != len(pipeline.node_id):
            warning=True
    if warning:
        print('Warning: length of id_nodes not equals length of coordinates per pipeline')
    countrydict=CountryPolyDict(TM_World_Borders_file)
    pipenodes=[]
    for pipe in pipelines:
        old_nodes = list(zip(pipe.lat,pipe.long))
        lat,long = pipeline_sequencing(pipe.lat,pipe.long,max_length=max_distance)
        new_nodes = list(zip(lat,long))
        old_ID_generator = old_node_IDs(pipe.node_id)
        new_node_ids = []
        first_id,second_id = next(old_ID_generator)
        new_ID_generator=new_node_ID(first_id,second_id)
        for i, new_node in enumerate(new_nodes[:-1]):
            if new_node in old_nodes:
                new_node_ids.append(first_id)
            else: new_node_ids.append(next(new_ID_generator))


            if new_nodes[i+1] in old_nodes:
                if i+1<(len(new_nodes)-1):
                    first_id,second_id = next(old_ID_generator)
                    new_ID_generator=new_node_ID(first_id,second_id)
        new_node_ids.append(second_id)
        pipe.node_id=new_node_ids
        pipe.lat=lat
        pipe.long=long
        old_countrycodes=pipe.country_code
        pipe.country_code=GetCountry4List(TM_World_Borders_file,long,lat,countrydict,old_countrycodes)
    if verbose==True:
        print('--Segmenting INET--')

    return pipelines



import json
import os
import time


def get_param(param_name,elements,i):
    """
    returns values of param_name of the Netz.elements[i]
    """
    res=elements[i].__dict__.get(param_name)    
    return res

def get_coordinates(elements,i):
    """
    returns coordinates in geoJSON format
    """
    lats=elements[i].__dict__.get('lat')    
    longs=elements[i].__dict__.get('long')  
    coordinates=[]
    
    "for line elements"
    if isinstance(longs,list):
        for coords in zip(longs,lats):
            coordinates.append(list(coords))     
        "for point elements"        
    else:
        coordinates=[longs,lats]        
    return coordinates
            
def create_GeoJson_Lines(elements):    
    """
    create GeoJSON output if Netz.elements consists of lines
    """    
    feature=[]
    for i in range(len(elements)):
        
        feature.append(       {'type'        : 'Feature',
                               'geometry'    : 
                              {'type'        : 'LineString',
                               'coordinates' : get_coordinates(elements,i)},
                               'properties'  : 
                              {'name'        : get_param('name',elements,i), 
                               'id'          : get_param('id',elements,i),
                               'country_code': get_param('country_code',elements,i) ,
                               'tags'        : get_param('tags',elements,i),
                               'param'       : get_param('param',elements,i),
                               'method'      : get_param('method',elements,i)}
                              })   
    return feature
    
def create_GeoJson_Points(elements):
    """
    create GeoJSON output if elements consists of points
    """    
    feature=[]
    for i in range(len(elements)):
        feature.append(       {'type'         : 'Feature',
                               'geometry'     : 
                              {'type'         : 'Point',
                               'coordinates'  : get_coordinates(elements,i)},
                               'properties'   : 
                              {'name'         : get_param('name',elements,i), 
                               'id'           : get_param('id',elements,i),
                               'country_code' : get_param('country_code',elements,i) ,
                               'tags'         : get_param('tags',elements,i),
                               'param'        : get_param('param',elements,i),
                               'method'       : get_param('method',elements,i)}
                               })   
    return feature
    
def geojson_export_element(Netz,Elementtyp,verbose=False):
    """
    Export Netz.Element to a geoJSON file
    """
    elements=Netz.__dict__.get(Elementtyp)
    #if verbose:
     #   print(elements)
      #  print(Elementtyp)
    if Elementtyp in (['PipeLines','PipeSegments']):
        output=json.dumps({'type': 'FeatureCollection',
          'features': create_GeoJson_Lines(elements)},indent=4)
    else:    
        output=json.dumps({'type': 'FeatureCollection',
          'features': create_GeoJson_Points(elements)},indent=4)
        
    return output

def export_geojson(Netz,netname,filepath='Output/GeoJSON/',verbose=False):
    """
    Export Netz object to separate geojson files
    """
    
    for pipe in Netz.PipeSegments:
        pipe.country_code=list(pipe.country_code)
    if verbose:
        print(os.path.dirname(filepath))    
        print('Filepathexist=',os.path.exists(os.path.dirname(filepath)))   
    if not os.path.exists(os.path.dirname(filepath)):
        os.makedirs(os.path.dirname(filepath))
        if verbose:
            print('created the following path: ', filepath)
        time.sleep(10)
        if verbose:
            print(os.path.exists(os.path.dirname(filepath)))
        
    if os.path.exists(os.path.dirname(filepath)):
        for elementname in Netz.CompLabels():   
            if verbose: print(elementname)
            if len(Netz.__dict__[elementname])>0:
                a=geojson_export_element(Netz,elementname,verbose)
                filename=netname+'_'+elementname+'.geojson'
                filename=os.path.join(filepath,filename)
                open(filename,"w+").write(a)
            
if __name__== '__main__':
     import M_Internet
     INET2=M_Internet.read(RelDirName = '../Eingabe/InternetDaten/')
     export_geojson(INET2,'INET',path='../Ausgabe/GeoJSON')



# -*- coding: utf-8 -*-
"""
Component Class
***************
"""

from . import M_DataAnalysis
import copy


def typcheck(x):
    if x == 'None':
        x = None

    if x is None:
        return x
    else:
        return float(x)
    pass



class Component(object):
    """Class **Component** with the following fixed attributes **id**, **name**, **source_id**, **node_id**,** country_code**,
	**lat**, **long**, and **comment**, and the following dicts, **tags**, **uncertainty**, **method**, **param**."""

    def __init__(self, id, name, source_id, node_id, country_code = None, lat = None, long = None, comment = None, tags = None, uncertainty = None, method = None, param = None):


        if lat is None:
            lat = None
        if long is None:
            long = None
        if tags is None:
            tags = {}
        if param is None:
            param = {}
        if uncertainty is None:
            uncertainty = {}
        if method is None:
            method = {}


        self.id             = id
        self.name           = name
        self.source_id      = source_id
        self.node_id        = node_id
        self.lat            = lat
        self.long           = long
        self.country_code   = country_code
        self.comment        = comment

        self.param          = param
        self.uncertainty    = uncertainty
        self.method         = method
        self.tags           = tags





    def copy2(self):
        """Method of creating a true independent copy of a class instance.
        """
        id          = self.id
        name        = self.name
        source_id   = self.source_id
        node_id     = self.node_id 
        
        RetElem      = Component(id = id, name = name, source_id = source_id, node_id = node_id)

        RetElem      = copy.deepcopy(self)

        return RetElem








    def get(self,value):
        return self.__dict__[value]

    def all(self,path=False):
        """Method to get info on component of Netz instancde"""
        for key in sorted(self.__dict__.keys()):
            param=self.__dict__[key]
            if type(param)!=dict:
                print(key + ': ' + str(self.__dict__[key]))
            if type(param)==dict:
                print(key + ': ')
                dict_list=list(param.items())
                for item in dict_list:         
                    if not (item[0] in ['path_lat','path_long']):
                        print('     '+ str(item))    
                    else:
                        #print('     '+ str(item[0]))
                        print('     '+ str(item[0])+':[',str(item[1][0]),'+ {'+str(len(item[1])-3),'values} +'+str(item[1][-1])+']')
                    


    def __repr__(self):
        return str(self.name)


    def AttribLables(self):
        """Method that returns list of string, that are the attribute labels for the component."""
        return ['id', 'name', 'source_id', 'node_id', 'lat', 'long', 'country_code', 'comment']




    def getPipeLength(self):
        length = 0
        if self.lat == None:
            length = None
        else:
            for idx in range(len(self.lat)-1):
                length = length + M_DataAnalysis.distance(self.long[idx], self.lat[idx], self.long[idx+1], self.lat[idx+1])

        return round(length*1000)/1000




    def get_Attrib(self, attribName, removeNone = False):
        """Method to get attribute values from all elements of this component.
        **attribName** are the inputs, and a list of values will
        be returned. None values can be removed, but can be suppressed with
        **removeNone**.

        \n.. comments:
        Input:
            attribName:         string of attribute name/label
            removeNone:         Boolean of None shall be removed or not
                                [Default = False]
        Return:
            attribList:         list of values
        """
        attribList = []

        for elem in self:

            if attribName in elem.__dict__['param'].keys():
                if isinstance(elem.param[attribName], list):
                    for tt in elem.param[attribName]:
                        attribList.append(tt)
                else:
                    attribList.append(elem.param[attribName])
            elif attribName in elem.__dict__.keys():
                if isinstance(elem.__dict__[attribName], list):
                    for tt in elem.__dict__[attribName]:
                        attribList.append(tt)
                else:
                    attribList.append(elem.__dict__[attribName])
            else:
                return []

        # leave data as is and return
        if removeNone == False:
            return attribList

        # removes Nones, and then return
        else:
            attribList2 = []
            for val in attribList:
                if val != None:
                    attribList2.append(val)
            return attribList2



class BorderPoints(Component):
    """ Component Class BorderPoints"""


class Compressors(Component):
    """ Component Class Compressors"""


class ConnectionPoints(Component):
    """ Component Class ConnectionPoints"""


class Consumers(Component):
    """ Component Class Consumers"""


class EntryPoints(Component):
    """ Component Class EntryPoints"""


class InterConnectionPoints(Component):
    """ Component Class InterConnectionPoints"""


class LNGs(Component):
    """ Component Class LNGs"""


class Nodes(Component):
    """ Component Class Nodes"""


class Operators(Component):
    """ Component Class Operators"""


class PipeLines(Component):
    """ Component Class PipeLines"""


class PipePoints(Component):
    """ Component Class PipePoints"""


class PipeSegments(Component):
    """ Component Class PipeSegments"""


class Productions(Component):
    """ Component Class Productions"""


class Storages(Component):
    """ Component Class Storages"""


class Processes():
    def __init__(self, Commments):
        self.Commments  = Commments


class PolyLine():
    def __init__(self, long, lat):
        self.long = long
        self.lat  = lat



class MetaData(object):
    def __init__(self):
        self.BorderPoints           = []
        self.Compressors            = []
        self.EntryPoints            = []
        self.InterConnectionPoints  = []
        self.LNGs                   = []
        self.PipePoints             = []
        self.Storages               = []










# #class Component(object):
# def makeLine(X = [], Y = []):

#     polyLine        = PolyLine(lat = [], long = [])
#     polyLine.long   = X
#     polyLine.lat    = Y


#    return polyLine

class OSMComponent(Component):
    def __init__(self, id, name, source_id, node_id, lat, long, country_code, tags,param,uncertainty,method):
        self.id             = id
        self.name           = name
        self.source_id      = source_id
        self.node_id        = node_id
        self.lat            = lat
        self.long           = long
        self.country_code   = country_code
        self.tags           = tags
        self.param = param
        self.uncertainty = uncertainty
        self.method = method
        self.comments = None
 
 

        # for key in param:
        #     if key!= 'uncertainty' and key != 'method' and key != 'source' and key != 'license':
        #         self.uncertainty[key]  = None
        #         self.method[key] = None













from . import M_GetCountry          as GC
#import Code.K_Component           as KC
from . import M_PCKL                as PCKL
from . import C_colors              as CC
import os, time

def Calc_OSM_Pipeline_CountryCodes(TM_World_Borders_file,
                                   PipeLine,countrycode):
    
    ''' Calulating lat/long for each pipeline point of a Pipeline '''
    
    countrypolydict=GC.CountryPolyDict(TM_World_Borders_file,
                                       predicted_countrycodes=countrycode)
    res=GC.GetCountry4List(TM_World_Borders_file,PipeLine.long,PipeLine.lat,countrypolydict=countrypolydict,predicted_countrycodes=countrycode)
    return res


def Calc_OSM_Pipelines_CountryCodes(TM_World_Borders_file,Pipelines,countrycode,verbose=False):
    '''
    Calculating lat/long for a list of pipelines
    '''
    pipelines_country_code_list=[]
    data_count=len(Pipelines)
    for i,pipeline in enumerate(Pipelines):
        countrycodes_of_line=Calc_OSM_Pipeline_CountryCodes(TM_World_Borders_file,
                                                             pipeline,countrycode)
        pipelines_country_code_list.append([pipeline.id,
                                            countrycodes_of_line])
        print(countrycodes_of_line)
        if verbose==True:
            print(f'{i+1}/{data_count}')
    return pipelines_country_code_list


def Create_Pipelines_CountryCodes(JSON_outputfile,TM_World_Borders_file,
                                  Netz,countrycode,verbose=False):
    '''
    Calculates pipeline countrycodes and stores it to file
    '''
    print(CC.Green+'Create countrycodes for pipelines:'+CC.End)
    pipeline_countrycodelist=Calc_OSM_Pipelines_CountryCodes(TM_World_Borders_file,
                                                             Netz.PipeLines,
                                                             countrycode,
                                                             verbose)
    print('\nPickle gas data to \n'+os.path.dirname(JSON_outputfile),'/pipeline_countrycode_list.pickle')
    PCKL.picklesavelist(pipeline_countrycodelist,
    'pipeline_countrycode_list.pickle',os.path.dirname(JSON_outputfile))
    pipesegment_countrycodelist=Calc_OSM_Pipelines_CountryCodes(TM_World_Borders_file,
                                                             Netz.PipeSegments,
                                                             countrycode,
                                                             verbose)
                                                             
    print('\nPickle gas data to \n'+os.path.dirname(JSON_outputfile),'/pipesegment_countrycode_list.pickle')
    PCKL.picklesavelist(pipesegment_countrycodelist,
    'pipesegment_countrycode_list.pickle',os.path.dirname(JSON_outputfile))
    pass
    
def load_pipeline_countrycode(JSON_outputfile):
    '''
    Loads Countrycodes from Json-File
    '''
    print("Load countrydata from ",JSON_outputfile)
    liste=PCKL.pickleloadlist(os.path.dirname(JSON_outputfile),
                              'pipeline_countrycode_list.pickle')
    
    return liste

def load_pipesegment_countrycode(JSON_outputfile):
    '''
    Loads Countrycodes from Json-File
    '''
    print("Load countrydata from ",JSON_outputfile)
    liste=PCKL.pickleloadlist(os.path.dirname(JSON_outputfile),
                              'pipesegment_countrycode_list.pickle')
    
    return liste

def Load_Pipelines_Countrycodes(JSON_outputfile,Netz,countrycode):
    '''
    loads countrycodes list for a Netz object and 
    
    '''
    pipe_count=len(Netz.PipeLines)
    countrycode_data=load_pipeline_countrycode(JSON_outputfile)
    data_count=len(countrycode_data)
    if pipe_count==data_count:
        for i in range(pipe_count):
            #print(i,':',countrycode_data[i][0],'vs',Netz.PipeLines[i].id )
            for j in range(data_count):
                if countrycode_data[j][0]==Netz.PipeLines[i].id:
                    Netz.PipeLines[i].country_code=countrycode_data[j][1]
                    for node_id,countrycode_value in zip(Netz.PipeLines[i].node_id,
                                                         Netz.PipeLines[i].country_code):
                        Netz.select('Nodes','id',node_id).country_code=countrycode_value
           
            
    else:
        print(CC.Warning+'2:Outdatet Countrypipeline file\n Please recreate countrycode with OSM_Pipelines_CountryCodes()'+CC.End)
        time.sleep(2)
    pass

def Load_Pipesegments_Countrycodes(JSON_outputfile,Netz,countrycode):
    '''
    loads countrycodes list for a Netz object and 
    
    '''
    pipe_count=len(Netz.PipeSegments)
    countrycode_data=load_pipesegment_countrycode(JSON_outputfile)
    data_count=len(countrycode_data)
    if pipe_count==data_count:
        for i in range(pipe_count):
            #print(i,':',countrycode_data[i][0],'vs',Netz.PipeSegments[i].id)
            for j in range(data_count):
                if countrycode_data[j][0]==Netz.PipeSegments[i].id:
                    Netz.PipeSegments[i].country_code=countrycode_data[j][1]
                    for node_id,countrycode_value in zip(Netz.PipeSegments[i].node_id,
                                                         Netz.PipeSegments[i].country_code):
                        Netz.select('Nodes','id',node_id).country_code=countrycode_value
        # else:
        # print(CC.Green+'Success: added countrycodes to ',data_count,' Pipelines'+CC.End)
        # time.sleep(2)
            
    else:
        print(CC.Warning+'2:Outdatet Countrysegment file\n Please recreate countrycode with OSM_Pipelines_CountryCodes()'+CC.End)
        time.sleep(2)
    pass


if __name__=='__main__':
    Create_Pipelines_CountryCodes(JSON_outputfile,OSM,'EU')
    Load_Pipelines_Countrycodes(JSON_outputfile,OSM,'EU')
    
    



# -*- coding: utf-8 -*-
"""
M_DataAnalysis
---------
"""

from math import sin, cos, sqrt, atan2, radians


def Pipelength(Pipelines):
    total=0
    for pipe in Pipelines:
        length      = routelength(pipe.long,pipe.lat)
        pipe.length = length
        total       = total+float(length)

    return total




def distance(lon1,lat1,lon2,lat2):
    ''' Calculate distance between 2 Points on Earth's surface.

    \n.. comments:
    Input:
        lon1: 		float of long pos 1
		lat1: 		float of lat pos 1
		lon2: 		float of long pos 2
		lat2: 		float of lat pos 2
    Output:
		distance 	float
		'''
    R = 6373.0 # radius earth
    lat1 = radians(lat1)
    lon1 = radians(lon1)
    lat2 = radians(lat2)
    lon2 = radians(lon2)

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = sin(dlat / 2)**2 + cos(lat1) * cos(lat2) * sin(dlon / 2)**2
    c = 2 * atan2(sqrt(a), sqrt(1 - a))

    distance = R * c

    return distance




def routelength(long, lat):
    """Returns total length of a pipeline (polyline, multi point pipeline)

    \n.. comments:
    Input:
        lon1: 		list of floats of long values
		lat1: 		lit of float of lat values
    Output:
		distance 	float

	"""

    length=0.0
    for i in (range(len(long)-1)):
        length = length + distance(long[i],lat[i],long[i+1],lat[i+1])

    return "{0:.3f}".format(length)




def set_lengths(elements):
    ''' Calculates length of Line Element and attach it with dict param

    \n.. comments:
    Sample:
           length(SciGrid.Pipelines) '''
    for element in elements:
        element.param['length_km']   = float(routelength(element.long, element.lat))
    pass




def sum_length(elements):
    """Sums values over all elements of the attribute length in dict param.

    \n.. comments:
    Input:
		elements: 	List of elements
    Output:
		sumlength:
	"""

    sumlength=0.0

    for element in elements:
        sumlength += float(element.param['length_km'])

    return sumlength







# -*- coding: utf-8 -*-
"""
M_PCKL
------
Module for pickle load and saving.

"""

import pickle
import os


def picklesavelist(element, filename, filedir):
    ''' Pickles list to file directory
	
	\n.. comments:
        Input: 
		    dictionary 
			filedirectory
    '''
    if not os.path.exists(filedir):
        os.makedirs(filedir)
    with open(os.path.join(filedir,filename), 'wb') as handle:
        pickle.dump(element, handle, protocol=pickle.HIGHEST_PROTOCOL)
        
    return




def pickleloadlist(filedirname,file):
    ''' UnPickles list from filedirectory
    '''

    with open(os.path.join(filedirname,file), 'rb') as handle:
        liste=pickle.load(handle)
        
    return liste




def picklesave(elements, filedir):
    ''' Saves network components into a Pickles file, where directory given through **filedir**.
	
	\n.. comments:
        Input: 
		    elements  		network component list
			filedir 		string of directory name
    '''
    if not os.path.exists(filedir):
        os.makedirs(filedir)
    for elementname  in elements:
        with open(os.path.join(filedir, elementname+'.pickle'), 'wb') as handle:
            pickle.dump(elements[elementname], handle, protocol=pickle.HIGHEST_PROTOCOL)
            
    return           




def pickleload(elements, filedir):
    ''' Restores dict of dicts which was previously stored by pickelsave()
	
	\n.. comments:
        Input: 
		    elements  		network component list
			filedir 		string of directory name
    '''
    for elementname  in elements:
        with open(os.path.join(filedir, elementname+'.pickle'), 'rb') as handle:
            elements[elementname] = pickle.load(handle)
            
    return elements





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''Library to convert OSM and CSV Objects to Plotable Items'''

import json
from unidecode import unidecode
from math import sin, cos, sqrt, atan2, radians

def distance(lon1,lat1,lon2,lat2):
    ''' Calulate distance between 2 Points on Earth's surface '''
    R = 6373.0 # radius earth
    lat1 = radians(lat1)
    lon1 = radians(lon1)
    lat2 = radians(lat2)
    lon2 = radians(lon2)

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = sin(dlat / 2)**2 + cos(lat1) * cos(lat2) * sin(dlon / 2)**2
    c = 2 * atan2(sqrt(a), sqrt(1 - a))

    distance = R * c
    return distance




def routelength(long, lat):
    length = 0.0

    if type(long) == float:
        length = 0
    else:
        for i in (range(len(long)-1)):
            length = length + distance(long[i], lat[i], long[i + 1], lat[i + 1])
    return "{0:.3f}".format(length)




def add_lengths(elements):
    ''' Caculates length of Line Element and attach it as parameter
    \n.. comments:
    Sample:
           length(SciGrid.Pipelines) '''
    for element in elements:
        element.param['length_km'] = routelength(element.long, element.lat)
    pass




def sum_length(elements):
    sumlength=0.0
    for element in elements:
        sumlength+=float(element.param['length_km'])
    return sumlength




def avg_lonlat(lonlat_array):
    long=0
    lat=0
    for entry in lonlat_array:
        long+=entry[0]
        lat+=entry[1]
    return float("{0:.6f}".format(long/(len(lonlat_array)))) ,float("{0:.6f}".format(lat/(len(lonlat_array))))




def Nodes2Points(elements, tagstyle = 1,parameter='',dictionary='param'):
    long    = []
    lat     = []
    taglist = []
    parameterlist=[]

    if parameter!='':
        for element in elements:
            parameterlist.append(element.__dict__[dictionary].get(parameter))

    # in case that long is a list
    if type(elements[0].long) == list:
        for idf, element in enumerate(elements):

                long.append(element.long[0])
                lat.append(element.lat[0])

                tag = '(' + str(element.id) + ')'+str(element.name) + '\n'
                if tagstyle >1:
                    tag+='ID:'+str(element.id)+'\n'
                    tag+='lat:' +str(element.lat[0])+'\n'
                    tag+='lon:' +str(element.long[0])+'\n'
                if tagstyle >2:
                    tag+=(json.dumps(element.__dict__[dictionary],ensure_ascii=False,indent=2)).replace('{','').replace('}','').replace('\\"','')
                taglist.append(tag)

    # in case that long is a float
    else:
        for idf, element in enumerate(elements):

            long.append(element.long)
            lat.append(element.lat)

            tag = '(' + str(element.id) + ') ' + str(element.name) + '\n'
            if tagstyle >1:
                tag+='ID:'+str(element.id)+'\n'
                tag+='lat:' +str(element.lat)+'\n'
                tag+='lon:' +str(element.long)+'\n'
            if tagstyle >2:
                tag+=(json.dumps(element.__dict__[dictionary],ensure_ascii=False,indent=2)).replace('{','').replace('}','').replace('\\"','')
            taglist.append(tag)


    if parameter!='':
        return long,lat,taglist,parameterlist
    else:
        return long,lat,taglist




def Ways2Lines(elements, min_length = 0, tagstyle = 1,parameter='',pipelinethickness=False,dictionary='param'):
    ''' convert Ways2Lines to plotable lines '''
    lines   = []
    long    = []
    lat     = []
    if len(elements)>0:
        for element in elements:
            lat         = element.lat
            long        = element.long
            if parameter!='':
                para=element.__dict__[dictionary].get(parameter,0)
            if pipelinethickness==True:
                thickness=element.__dict__[dictionary].get('diameter_mm')
            if type(lat) != type(None):
                tag         = 'ID:'+str(element.id)
                if parameter=='':
                    line        = [long,lat,tag]
                else:
                    line        = [long,lat,tag,para]
                if pipelinethickness==True:
                    line.append(thickness)
                linelength  = float(routelength(long,lat))
                if linelength>=min_length:
                    lines.append(line)
                else:
                    print('M_PlotObjects.Ways2Lines: Too short')
    return lines




def OSMnodes2points(elements):
    ''' convert OSMnodes to plotable Points '''
    long=[]
    lat=[]
    taglist=[]
    if "Node" in elements.keys():
        for entry in elements["Node"]:
            long.append(elements["Node"][entry]["lonlat"][0])
            lat.append(elements["Node"][entry]["lonlat"][1])
            tag=('ID:'+entry)
            tag+=('\n'+(unidecode(json.dumps(elements["Node"][entry]["tags"],ensure_ascii=False,indent=2))).replace('{','').replace('}','').replace('\\"',''))
            taglist.append(tag)
    return long,lat,taglist




def OSMways2points(elements,data,min_length=0):
    ''' convert OSMways  to Plotable Points '''
    long=[]
    lat=[]
    taglist=[]
    refs=[]
    long_avgs=[]
    lat_avgs=[]
    if "Way" in elements.keys():
        for entry in elements["Way"]:
            refs=elements["Way"][entry]['refs']
            lonlat_array=[]
            for ID in refs:
    #            for ID in reference:
                lonlat_array.append(data["Node"][str(ID)]['lonlat'])
            [long_avg,lat_avg]=avg_lonlat(lonlat_array)
            long=[]
            lat=[]

            #long
            linelength=0.0
            for coords in lonlat_array:
                long.append(coords[0])
                lat.append(coords[1])
            #lengths.append(routelength(long_way,lat_way))
            linelength=float(routelength(long,lat))
            if linelength>=min_length:
                long_avgs.append(long_avg)
                lat_avgs.append(lat_avg)
                tag='ID:'+entry
                tag+='\n'+'length:'+str(linelength)+' km \n'+unidecode(json.dumps(elements["Way"][entry]["tags"],ensure_ascii=False,indent=2)).replace('{','').replace('}','').replace('\\"','')
                taglist.append(tag)
    return long_avgs,lat_avgs,taglist




def OSMways2lines(elements,data,min_length=0):
    ''' Convert OSMways to plotable lines '''
    refs=[]
    lines=[]
    id=[]
    if "Way" in elements.keys():
        for entry in elements["Way"]:
            refs=elements["Way"][entry]['refs']
            lonlat_array=[]
            for ID in refs:
                lonlat_array.append(data["Node"][str(ID)]['lonlat'])
            long=[]
            lat=[]
            line=[]
            length=[]
            for coords in lonlat_array:
                long.append(coords[0])
                lat.append(coords[1])
            linelength=float(routelength(long,lat))
            id.append(entry)
            length.append(linelength)
            tag='ID:'+entry+'\n'+'length:'+(routelength(long,lat))+' km \n'+unidecode(json.dumps(elements["Way"][entry]["tags"],ensure_ascii=False,indent=2)).replace('{','').replace('}','').replace('\\"','')
            line=[long,lat,tag]
            if linelength>=min_length:
                lines.append(line)
    return lines




def OSMrelations2lines(elements,data):
    ''' Convert OSMrelation to plotable lines '''
    refs=[]
    lines=[]
    for entry in elements["Relation"]:
        refs=elements["Relation"][entry]['refs']
        lonlat_array=[]
        for ID in refs:
            lonlat_array.append(data["Node"][str(ID)]['lonlat'])
        long=[]
        lat=[]
        line=[]
        for coords in lonlat_array:
            long.append(coords[0])
            lat.append(coords[1])
        tag='ID:'+entry+'\n'+unidecode(json.dumps(elements["Way"][entry]["tags"],ensure_ascii=False,indent=2)).replace('{','').replace('}','').replace('\\"','')
        line=[long,lat,tag]
        lines.append(line)
    return lines




def create_CSVPlotPoints(elementlist,nodelist):
    ''' Convert CSV Objects to plotable points '''
    long=[]
    lat=[]
    tag=[]
    for element in elementlist:
        for node in nodelist:
            if element.id == node.id:
                long.append(node.long)
                lat.append(node.lat)
                tag.append(node.id)
    #plotobj=[long,lat,tag]
    return long,lat,tag




def create_CSVPlotLines(linelist,nodelist):
    '''Convert CSV Objects to plotable lines '''
    long=[]
    lat=[]
    tag=[]
    linename=''
    pipes=[]
    for line in linelist:
        if line.name != linename:
            pipe=[long,lat,linename]
            pipes.append(pipe)
            linename=line.name
            long=[];lat=[];tag=[]
        for node in nodelist:
            if line.node_id == node.id:
                long.append(node.long)
                lat.append(node.lat)
                tag.append(node.id)
    pipes.append(pipe)
    return pipes










import array    		    as arr

def find_pos_StringInList(String, ListOfStrings, Str2Lower = False):
    """Find positions, where string is equal to a string in a list.  Input are **String**, and **ListOfStrings**
    and boolean **Str2Lower**.
	
    \n.. comments: 
    Input:
        String:          	String, 
        ListVonStrings:  	List of strings 
		Str2Lower: 			Boolean re converting strings to lower.
							(default = False)
    Return:
        pos:             	An array of type Integer, containing positions, where String were found in ListOfStrings."""

    # Initializierung von Variabeln
    pos = arr.array('i', [])
    
    if Str2Lower == False:
        if type(String) == str:
            count = 0
            
            for Name in ListOfStrings:
                if String == Name:
                    pos.append(count)
                count = count + 1
                
        elif type(String) == list:
            
            for Str in String:
                count = 0
                for Name in ListOfStrings:
                    if Str == Name:
                        pos.append(count)
                    count = count + 1
        else:
            print('ERROR: M_FindPos.find_pos_StringInList: Code noch NICHT geschrieben 1')
    else:
        if type(String) == str:
            count = 0
            String = String.lower()
            
            for Name in ListOfStrings:
                if Name is not None:
                    if String == Name.lower():
                        pos.append(count)
                count = count + 1
                
        elif type(String) == list:
            
            for Str in String:
                count = 0
                Str = Str.lower()
                for Name in ListOfStrings:
                    if Name is not None:
                        if Str == Name.lower():
                            pos.append(count)
                    count = count + 1
        else:
            print('ERROR: M_FindPos.find_pos_StringInList: Code noch NICHT geschrieben 2')
        
    return pos

def makeList(stringVal):
    """Function to convert a string values into a list"""

    listVal = []
    if stringVal != None:
        if isinstance(stringVal, str):
            if '[\'' in stringVal and '\']' in stringVal:
                stringVal = stringVal[1:-2]
                stringVal   = stringVal
                stringVal   = stringVal.replace('\', \'', '\',\'')
                for val in stringVal.split(','):
                    if val[0] == '\'':
                        val = val[1:]
                    if val[-1] == '\'':
                        val = val[:-1]
                    listVal.append(val)
            else:
                listVal = stringVal
        else:
            listVal = stringVal

    else:
        listVal = stringVal

    return listVal



# -*- coding: utf-8 -*-
"""
Net class
*********
"""

from __future__ import print_function
from . import K_Component as K_Component


import copy
import ast

class NetComp(object):
    """Main class **NetComp**. Class containing components such as **BorderPoints** or
    **Compressors**, used for the non-OSM data set.
    """
    def __init__(self):
        self.SourceName             = ['']
        self.BorderPoints           = []    # BP
        self.Compressors            = []    # CO
        self.ConnectionPoints       = []    # CP
        self.Consumers              = []    # CS
        self.EntryPoints            = []    # EP
        self.InterConnectionPoints  = []    # IC
        self.LNGs                   = []    # LG
        self.Nodes                  = []    # NO
        self.PipePoints             = []    # PP
        self.PipeSegments           = []    # PS
        self.PipeLines              = []    # PL
        self.Productions            = []    # PD
        self.Storages               = []    # SR

        self.Processes              = []



    def copy2(self):
        """Method of creating a true independent copy of a class instance.
        """
        RetNetz      = NetComp()

        RetNetz      = copy.deepcopy(self)

        return RetNetz




    def AttribLables(self):
        return ['id', 'name', 'source_id', 'node_id', 'lat', 'long', 'country_code', 'comment']




    def CompLabelsSpot(self):
        return ['Compressors', 'Consumers', 'ConnectionPoints', 'InterConnectionPoints',
                'BorderPoints', 'LNGs', 'Storages', 'EntryPoints', 'Productions', ]




    def CompLabelsPipes(self):
        return ['PipeSegments', 'PipeLines']





    def CompLabelsNoNodes(self):
        return [*self.CompLabelsSpot(), *self.CompLabelsPipes(),
                'PipePoints']




    def CompLabels(self):
        return ['Nodes', *self.CompLabelsNoNodes()]


    def copy(self):
        """Method to create copy of instance.
        """

         # BookKeeping
        self.Processes.append(K_Component.Processes('K_Netze.Netz..copy: Creation of copy of  Netze instance'))

        # Initialization
        self2 = NetComp()
        try:
            for key in sorted(self.__dict__.keys()):
                if self.__dict__[key] is not None:
                    for dd  in self.__dict__[key]:
                        self2.__dict__[key].append(dd)
        except:
            pass
        return self2


    def getcountry4pipelines(self):
        """Method of getting pipeline CountryCode from Nodes-list exists
        """

        for i in range(len(self.PipeLines)):
            countrycodelist=[]
            for id in self.PipeLines[i].node_id:
                countrycode=self.Nodes[ast.literal_eval(repr(self.Nodes)).index(id)].country_code
                countrycodelist.append(countrycode)
            self.PipeLines[i].country_code=countrycodelist
        print('Got countrycodes for pipelines from nodes list')
        pass




    def para(self, element,parameter):
        """ Returns valuelist of a p from an element type
        """
        liste=[]
        print(element,'-',parameter,':')
        for single_element in self.__dict__[element]:
            liste.append(single_element.param.get(parameter))
        return liste


    def get_elevations(self):
        """Method uses bing api to get elevation of Nodes and write them to param dictionary elevation_m
        """
        from Code.M_Elevation import get_elevation
        nodelist=[]
        for node in self.Nodes:
            nodelist.append([node.lat,node.long])
        elevationlist=get_elevation(nodelist)
        for elevationVal, node in zip(elevationlist,self.Nodes):
            node.param.update({'elevation_m':elevationVal})
            node.method.update({'elevation_m':'bing API'})
            node.uncertainty.update({'elevation_m': 0})
        pass



    def rename_nodes(self,prestring='LKD_'):
        """Renaming of Node ids in Network
        """
        newnodelist=[]
        complist=self.CompLabels()
        for component in complist:
            for element in self.__dict__[component]:
                newnodelist=[]
                for node in element.node_id:
                    newnodelist.append(prestring+node)
                element.node_id=newnodelist

        complist=self.CompLabelsSpot()+['Nodes',]
        for component in complist:
            for element in self.__dict__[component]:
                element.id=(prestring+element.id)
        pass


    def all(self,component='',param='',verbose=True):
        """Method of displaying of all attributes from the NetComp class instance.
        """

        # checking how many components have more than zero element
        for key in self.__dict__.keys():
            # print(key)
            # if len(self.__dict__[key]) > 0:
            CompCount =  len(self.__dict__[key])
        if verbose==True:
            print("--------------------------------------")
            print("{0:30s} {1:>6s}".format('Source ', str(self.SourceName[0])))
            print("{0:30s} {1:>6s}".format('total component type count',str(CompCount)))
            print("--------------------------------------")
            for key in sorted(self.__dict__.keys()):
                if key == 'SourceName':
                    pass
                else:
                    print("{0:30s} {1:>6s}".format(key, str(len(self.__dict__[key]))))

            print("--------------------------------------")
            print("{0:30s} {1:>6s}".format('Length of PipeLines    [km]', str(round(self.sumLength('PipeLines')))))
            print("{0:30s} {1:>6s}".format('Length of PipeSegments [km]', str(round(self.sumLength('PipeSegments')))))
        if (component and param) !='':
            tagliste=[]
            for element in self.__dict__[component]:
                if verbose==True:
                    print(element.__dict__[param])
                tagliste.append(element.__dict__[param])
            return tagliste
        
            
        


    def sumLength(self, compName = 'PipeLines'):
        """Method returning total sum length of all pipeline/pipeSegment in units of [km]."""

        RetSum = 0

        for pipe in self.__dict__[compName]:
            if 'length_km' in pipe.param:
                if pipe.param['length_km'] != None:
                    RetSum = RetSum + float(pipe.param['length_km'])
        return RetSum


    def select2(self,CompName,AttribName,AttribVal):
        liste=[]
        for item in self.__dict__[CompName]:
            if AttribVal in item.__dict__[AttribName]:
                liste.append(item)
                #self2.__dict__[CompName].append(item.__dict__[AttribName])
        return liste


    def select_with_exception(self,CompName,AttribName,AttribVal,exceptionlist):
        liste=[]
        for item in self.__dict__[CompName]:
            if AttribVal in item.__dict__[AttribName] or item.id in exceptionlist:
                liste.append(item)
                #self2.__dict__[CompName].append(item.__dict__[AttribName])
        return liste

    def select(self,CompName,AttribName,AttribVal):
        
        for item in self.__dict__[CompName]:
            if AttribVal == item.__dict__[AttribName]:
                return item
                #self2.__dict__[CompName].append(item.__dict__[AttribName])
                #yield item

    def reduce(self, AttribVal, AttribName = 'country_code'):
        '''Method of reducing data to a country specified by AttribVal=country_code  
        \n.. comments:
            Input:
				AttribVal: 		Attribute value
				AttribName: 	String of attribute name
								(default = 'country_code')'''
		
        self2 = self.copy()
        #reduce all PipeLines and PipeSegments

        if AttribVal=='EU' and AttribName=='country_code':
            return self2
        else:
            CompNames=self.CompLabelsPipes()
            
            nodes_to_keep=[]
            for compname in CompNames:
                self2.__dict__[compname]=self.select2(CompName=compname,
                                                      AttribName = AttribName,
                                                      AttribVal = AttribVal)
                                                      

                for comp in self2.__dict__[compname]:
                    nodes_to_keep+=comp.node_id

            # create ID List of points to keep

            PointCompNames = self.CompLabels()
            for compname in PointCompNames:  
                self2.__dict__[compname]=self.select_with_exception(CompName=compname,
                                                      AttribName = AttribName,
                                                      AttribVal = AttribVal,
                                                      exceptionlist=nodes_to_keep)       
      
                    
                                                      
            print('Reduce data to Country:', AttribVal)
            return self2


    def select_from_param(self,CompName,AttribName,AttribVal):
        for item in self.__dict__[CompName]:
            if AttribVal == item.param[AttribName]:
                return item

        return False

    def path2coords(self,CompName):
        for pipe in self.__dict__[CompName]:
            pipe.lat=pipe.param['path_lat']
            pipe.long=pipe.param['path_long']
        pass

    def Emap_path2coords(self,CompName):
        for pipe in self.__dict__[CompName]:
            pipe.lat=pipe.lat[0],*pipe.param['path_lat'],pipe.lat[-1]
            pipe.long=pipe.long[0],*pipe.param['path_long'],pipe.long[-1]
        pass


    def select_from_Pipelines(self,CompName,AttribName,AttribVal):
        for item in self.__dict__[CompName]:
            if AttribVal in item.__dict__[AttribName]:
                pos=item.__dict__[AttribName].index(AttribVal)
                point={}
                point.update({'id':AttribVal})
                point.update({'lat':item.__dict__['lat'][pos]})
                point.update({'long':item.__dict__['long'][pos]})
                point.update({'country_code':item.__dict__['country_code'][pos]})
                return point
        return False

    def search(self,CompName,AttribName,AttribVal):
        for item in self.__dict__[CompName]:
            if AttribVal in item.__dict__[AttribName]:
                return item


    def getElem(self,CompName,AttribName,AttribVal):
        """Returns the element for which a requested attribute value could be found
        """
        for item in self.__dict__[CompName]:
            if AttribVal in item.__dict__[AttribName]:
                return item
        return None


    def Lines2Nodes(self,Line='PipeLines',Node="Nodes"):
        """ Creating Point from Line Element
        """
        def non_existing_nodes(Nodelist,Newnodes):
            non_existing_nodes=[]
            for newnode in Newnodes:
                flag=False
                for node in Nodelist:
                    if node.id==newnode.id:
                        flag=True
                if flag==False:
                    non_existing_nodes.append(newnode)
            return non_existing_nodes

        nodes=[]
        # Checking for the first two PipePoints
        for pipe in self.__dict__[Line]:
            for i,node in enumerate(zip(pipe.lat,pipe.long)):
                print(i,pipe.id,sep=' ')
                nodes.append(K_Component.__dict__['Nodes'](
                          id        = pipe.node_id[i],
                          name      = pipe.node_id[i],
                          source_id = [pipe.id],
                          node_id   = pipe.node_id[i],
                          country_code = pipe.country_code[i],
                          lat       = node[0],
                          long      = node[1]))
        #Only add non existing nodes
        for node in non_existing_nodes(self.__dict__[Node],nodes):
            self.__dict__[Node].append(node)
        pass



    def PipeLines2PipeSegments(self):
        """Method of converting PipeLines to PipeSegments.
        And changing length value of PipeSegments
        """

        RetPipeSegments = []

        # Checking for the first two PipePoints
        for pipe in self.PipeLines:
            if len(pipe.node_id) == 2:
                RetPipeSegments.append(K_Component.PipeSegments(id = pipe.id,
                                name        = pipe.name,
                                source_id   = pipe.source_id,
                                node_id     = pipe.node_id,
                                lat         = pipe.lat,
                                long        = pipe.long,
                                country_code = pipe.country_code,
                                param       = pipe.param.copy()))
            else:
                for ii in range(len(pipe.node_id) - 1):
                    if pipe.lat == None:
                        RetPipeSegments.append(K_Component.PipeSegments(id = pipe.id + "_EE_" + str(ii),
                                name        = pipe.name+str(ii),
                                source_id   = pipe.source_id,
                                lat         = None,
                                long        = None,
                                node_id     = pipe.node_id[ii : ii+2],
                                country_code = pipe.country_code,
                                param       = pipe.param.copy()))
                    else:
                        print('ERROR: K_Netze.PipeLines2PipeSegments: code not written yet, as lat long missing.')
                        RetPipeSegments.append(K_Component.PipeSegments(id = pipe.id + "_EE_" + str(ii),
                                name        = pipe.name+str(ii),
                                source_id   = pipe.source_id,
                                node_id     = pipe.node_id[ii : ii+2],
                                country_code = pipe.country_code,
                                param       = pipe.param.copy()))


        self.PipeSegments = RetPipeSegments
        self.replace_length(compName = 'PipeSegments')

        #return []


    def PipeSegments2PipePoints(self):
        """ Method of converting PipeSegments to PipePoints
            Not working probably, lat, long!!!
        """

        RetPipePunkte = []
        count = 0
        # Checking for the first two PipePoints
        for pipe in self.PipeSegments:
            RetPipePunkte.append(K_Component.PipePoints(id = pipe.id,
                          name      = pipe.name,
                          source_id = pipe.source_id,
                          node_id   = [pipe.node_id[0]],
                          country_code = pipe.country_code,
                          lat       = None,
                          long      = None))
            RetPipePunkte.append(K_Component.PipePoints(id = pipe.id,
                          name      = pipe.name,
                          source_id = pipe.source_id,
                          node_id   = [pipe.node_id[1]],
                          country_code = pipe.country_code,
                          lat = None,
                          long = None))
            count = count + 1
        self.PipePoints = RetPipePunkte





class NetComp_OSM(NetComp):
    """
    same as NetComp but more elements, [outdated]
    """
    def __init__(self):
        self.SourceName             = ['']
        self.BorderPoints           = []    # BP
        self.Compressors            = []    # CO
        self.Compressors_Lines      = []    # CO
        self.ConnectionPoints       = []    # CP
        self.Consumers              = []    # CS
        self.EntryPoints            = []    # EP
        self.InterConnectionPoints  = []    # IC
        self.LNGs                   = []    # LG
        self.Nodes                  = []    # NO
        self.PipePoints             = []    # PP
        self.PipeSegments           = []    # PS
        self.PipeLines              = []    # PL
        self.Markers                = []
        self.SeaMarkers             = []
        self.Productions            = []    # PD
        self.Storages               = []    # SR
        self.Processes              = []


    pass









# -*- coding: utf-8 -*-
import sys
import pathlib
import setuptools 

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setuptools.setup(
    name="osmscigrid",
    version="0.0.13",
    description="Extracting gas transport pipelines OSM pbf-files and exporting to SciGRID_gas structure",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/osmscigrid",
    author="Adam Pluta",
    author_email="Adam.Pluta@dlr.de",
    package_dir={'':'src'},
    packages=setuptools.find_namespace_packages(where='src'),
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
     python_requires='>=3.6',
     install_requires=['protobuf >= 3, < 4', 
                       'esy-osm-pbf >= 0',
                       'Unidecode==1.1.1', 
                       'pandas>=0.25.3',
                       'stemming',
                       'pathlib>=1.0.1',
                       'esy-osmfilter',
                       'quantulum3==0.7.6',
                       'mplcursors==0.4',
                       'descartes==1.1.0',
                       'adjustText==0.7.3',
                       'numpy==1.19.4',
                       'PyShp',
                       'geopy',
                       'Shapely==1.7.1']
            
)



#from geopy.distance import geodesic
from   scipy.spatial.distance       import cdist
    
def find_nearest(coords,path_coords):
    distance_matrix=cdist(coords,path_coords)
    return distance_matrix


def path_segmenting(coords,path_coords):
    distance_matrix=find_nearest(coords,path_coords)
    index=[distance_row.argmin() for distance_row in distance_matrix]
    index_pairs=tuple(zip(index,index[1:]))
    #print(path_coords)
    paths=[path_coords[pair[0]:pair[1]+1] for pair in index_pairs]
    return paths[0]



if __name__ == '__main__':

    lat_path=[0,0.1,0.5,1,1.5,2,3]
    long_path=list(lat_path)
    path_coords=list(zip(lat_path,long_path))
    lat=[0,1]
    long=list(lat)
    coords=list(zip(lat,long))
    distance_matrix=find_nearest(coords,path_coords)
    result=path_segmenting(coords,path_coords)
    #print(result)
    



"""Read funktion: OSM Data (pipelines) into K_Net Class"""

#######################################################################
#All functions are called from the makefile                           #
#######################################################################

#from Code.OSM.M_OSM_CreateElements import M_OSM_CreateElements
#from unidecode                     import unidecode
#import Code.OSM.C_colors              as CC
import esy.osmfilter as EO
from . import M_OSM_convert 
from esy.osmfilter import Node,Way,Relation
from esy.osmfilter import osm_colors as CC


def read(PBF_inputfile,
         JSON_outputfile,
         TM_World_Borders_file,
         countrycode='EU', 
         NewPreFilterData=True, 
         CreateElements=True, 
         LoadElements=True,
         CreateCountryCodeForLines=True, 
         LoadCountryCodeForLines=True,
         length=100,
         min_length=0,
         multiprocess=True,
         verbose=False):
    """ Reads the OSM PipelineNetwork into K_Net Class
    \n.. comments: 
    Call: 
        
         OSM_Netz=M_OSM.read(Info=InfoOSM1)
         CreateElements = True (If the countrydata are used for the first time)
         NewPreFilterData = True (If the countrydata are used for the first time) 
         
    Return Data:
         element
         
         if elements are ways and than you can finde the referenced nodes in Data
         
         eg. to access lon lat for the node '6037838916' 
         Data['Node']['6037838916']['lonlat']"""
    
 
    

    prefilter = {
        Node: {"substance"    : ["gas","cng"],
                      "pipeline"     : ["substation","marker","valve",
                                        "pressure_control_station",True],
                      "product"      : ["gas","cng"],
                      "substation"   : ["valve","measurement", "inspection_gauge",
                                        "compression","valve_group"],
                      "man_made"     : ["pipeline","pipeline_marker","pumping_station"],
                      "gas"          : ["station",True],
                      "content"      : ["gas","cng"],
                      "storage"      : ["gas","cng"],
                      "industrial"   : ["gas","terminal","wellsite","cng"]
                      },

        Way: {"substance"     : ["gas","cng"],
                     "man_made"      : ["gasometer","pipeline","petroleum_well",
                                        "pumping_station", "pipeline_marker",
                                        "pipeline_station","storage_tank",
                                        "gas_cavern"],
                     "industrial"    : ["gas","terminal","cng"],
                     "gas"           : ["station"],
                     "content"       : ["gas","cng"],
                     "pipeline"      : ["substation","marker","valve",
                                        "pressure_control_station",True],
                     "storage"       : ["gas","cng"],
                     "seamark:type"  : ["pipeline_submarine"],
                     "land_use"      : ["industrial:gas"]
                      },

        Relation: {"substance"    : ["gas","cng"],
                      "man_made"     : ["gasometer","pipeline","petroleum_well",
                                        "pumping_station",
                                       "pipeline_marker","pipeline_station",
                                       "storage_tank","gas_cavern"],
                      "industrial"   : ["gas","terminal","cng"],
                      "gas"          : ["station"],
                      "content"      : ["gas","cng"],
                      "pipeline"     : ["substation","marker","valve",
                                        "pressure_control_station",True],
                      "storage"      : ["gas","cng"],
                      "seamark:type" : ["pipeline_submarine"],
                      "land_use"     : ["industrial:gas"]
                      }
        }

    whitefilter=[(("man_made","pipeline"),("substance","gas")),
                (("man_made","pipeline"),("substance","cng")),
                (("man_made","pipeline"),("substance","natural_gas")),
                (("man_made","pipeline"),("pipeline,type","natural_gas")),
                (("man_made","pipeline"),("type","gas")),
                (("man_made","pipeline"),("type","natural_gas")),  
                (("man_made","pipeline"),("type","cng")), 
                (("man_made","pipeline"),("industrial","gas")), 
                (("man_made","pipeline"),("industrial","cng")), 
                (("man_made","pipeline"),("industrial","natural_gas")),]
    
    blackfilter     =[("pipeline","substation"),
                    ("substation","distribution"),
                    ('usage', 'distribution'),
                    ('pipeline:type','water'),
                    ('pipeline:type', 'sewer'),
                    ("pumping_station","water"),
                    ("pumping_station","sewage"),
                    ('pumping_station','wastewater'),
                    ("type","wastewater"),
                    ("type","fuel"),
                    ("type","sewage"),
                    ("type","oil"),
                    ("type","water"),
                    ("substance","sewage"),
                    ("substance","water"),
                    ("substance","hot_water"),
                    ("substance","fuel"),
                    ("substance","wastewater"),
                    ("substance","rainwater"),
                    ("substance", "drain"),
                    ("substance", "heat"),
                    ("substance", "gas,heat"),
                    ("substance", "heat,gas"),
                    ("substance", "ammonia"),
                    ("substance", "ethylen"),
                    ("substance","oil")]
#    blackfilter = [("pipeline","substation"),]
#    whitefilter =[(("man_made","pipeline"),("waterway","drain")),]
    if verbose:
        print(CC.Caption+' Creating OSM net'+CC.End)
        print(f'Create:{CreateElements}')
    

    
    [Data,Elements]=EO.run_filter('pipelines',PBF_inputfile, JSON_outputfile, 
                                  prefilter, whitefilter, blackfilter, 
                                  NewPreFilterData=NewPreFilterData, 
                                  CreateElements=CreateElements, 
                                  LoadElements=LoadElements,verbose=verbose, 
                                  multiprocess=multiprocess)
#    print(Data)
     
#    [Data,Elements] = M_OSM_CreateElements('pipelines',countrycode,PBF_inputfile,
#                          JSON_outputfile, NewPreFilterData=NewPreFilterData, 
#                          CreateElements=CreateElements, LoadElements=LoadElements)
    ### Hier zwischenergebnisse speichern
    # min_length=min
    # length=length
    OSM=M_OSM_convert.RawData2ClassData(Data,Elements,JSON_outputfile,TM_World_Borders_file,
                                        countrycode=countrycode, 
                                        CreateCountryCodeForLines=CreateCountryCodeForLines,
                                        LoadCountryCodeForLines=LoadCountryCodeForLines,
                                        min_length=min_length,segment_length=length,verbose=verbose)
    

    return OSM

#def createall():
#    import configparser
#    from  configparser import  ExtendedInterpolation
#    from  pathlib            import Path
#    import os
#    import time
#    CountryCodes=["AL","XK","AM","LV","AT","LI","AZ","LT","BY","LU","BE","MT","BA",
#          "MD","BG","ME","HR","NL","CY","NO","CZ","PL","PT","EE","RO","FI",
#          "RS","FR","SK","GE","SI","DE","ES","EL","SE","HU","CH","IS","TR",
#          "IE","UA","IT","GB","DK","RU","EU"]
#
#    InfoOSM     = configparser.ConfigParser(interpolation=ExtendedInterpolation())
#    Setup_OSM      = Path(os.getcwd() + '/Setup/Setup_OSM.ini')
#    InfoOSM.read(Setup_OSM)
#
#    for countrycode in CountryCodes:
#    #    print('OSM_'+entry+'_ES2050')
#        InfoOSM1        = InfoOSM['OSM_'+countrycode+'_local']
#        #InfoOsmosis = InfoOSM['OSMOSIS_ADAM']
#        PBF_inputfile   = InfoOSM1['PBF_inputfile']
#        JSON_outputfile = InfoOSM1['JSON_outputfile']
#
#        print("PBF_inputfile for "+countrycode+" exists : "+ str(os.path.isfile(PBF_inputfile)))
#        if (os.path.isfile(PBF_inputfile)):
#            print(time.strftime('%m/%d/%Y', time.gmtime(os.path.getmtime(PBF_inputfile))))
#            print('check')
#            read(InfoOSM1,countrycode = countrycode, NewPreFilterData = True, CreateElements  = True,LoadElements = True, CreateCountryCodeForLines = True, LoadCountryCodeForLines=True)
#        else:
#            print('pbf-file not found')
        
        
        
#        [GasData,elements]=M_OSM_CreateElements(countrycode,PBF_inputfile,JSON_outputfile, CreateNewGasData=True, CreateElements=True,LoadElements=False,CreateCountryCodeForPipelines = True)
#    ########
#    #Markers
#    ########
#    [nodes,ids,tags,node_ids,tagdicts] = OSMnodes2net(elements["markers"],GasData)
#    for i,entry in enumerate(zip(nodes,ids,tags,node_ids)):
#        Marker=KC.OSMComponent(id = ids[i], name=ids[i], node_id=node_ids[i],source_id = ['OSM'], lat = nodes[i][1], 
#                              long = nodes[i][0], country_code = countrycode,tags = tagdicts[i])
#        Ret_Data.Markers.append(Marker)
#        
#    ########
#    #SeaMarkers
#    ########
#    [lines,ids,lengths,tags,node_ids,tagdicts]= OSMways2net(elements["seamarkers"],GasData,min_length=0)
#    for i,entry in enumerate(zip(lines,ids,lengths,tags)):
#        SeaMarker=KC.OSMComponent(id = ids[i], name=ids[i], node_id=node_ids[i],source_id = ['OSM'], lat = lines[i][1], 
#                              long = lines[i][0], country_code = countrycode, length=lengths[i],tags = tagdicts[i])
#        Ret_Data.SeaMarkers.append(SeaMarker)
            
        
#    ########
#    #Compressors_lines
#    ########
#    [lines,ids,lengths,tags,node_ids,tagdicts] = OSMways2net(elements["compressors_lines"],GasData,min_length=0)
##    maxlines=len(lines)
#    for i,entry in enumerate(zip(lines,ids,lengths,tags)):
#        Compressor_Lines=KC.OSMComponent(id = ids[i], name=ids[i], node_id=node_ids[i],source_id = ['OSM'], lat = lines[i][1], 
#                              long = lines[i][0], country_code = countrycode,tags = tagdicts[i], length=lengths[i])
#        Ret_Data.Compressors_Lines.append(Compressor_Lines)
#        
#    ########
#    #Compressors
#    ########
#    [nodes,ids,tags,node_ids,tagdicts] = OSMnodes2net(elements["compressors"],GasData)
#    for i,entry in enumerate(zip(nodes,ids,tags,node_ids)):
#        Compressor=KC.OSMComponent(id = ids[i], name=ids[i], node_id=node_ids[i],source_id = ['OSM'], lat = nodes[i][1], 
#                              long = nodes[i][0], country_code = countrycode,tags = tagdicts[i])
#        Ret_Data.Compressors.append(Compressor)
#
#    
#    
#    ########
#    #Pipeline3
#    ########
#    [lines,ids,lengths,tags,node_ids,tagdicts] = OSMways2net(elements["pipelines3"],GasData,min_length=0)
#    for i,entry in enumerate(zip(lines,ids,lengths,tags)):
#        PipeLine3=KC.OSMComponent(id = ids[i], name=ids[i], node_id=node_ids[i],source_id = ['OSM'], lat = lines[i][1], 
#                              long = lines[i][0], country_code = countrycode, length=lengths[i],tags = tagdicts[i])
#        Ret_Data.PipeLines3.append(PipeLine3)
#
#    
#    
#    ########
#    #Pipeline2
#    ########
#    [lines,ids,lengths,tags,node_ids,tagdicts] = OSMways2net(elements["pipelines2"],GasData,min_length=0)
#    for i,entry in enumerate(zip(lines,ids,lengths,tags)):
#        PipeLine2=KC.OSMComponent(id = ids[i], name=ids[i], node_id=node_ids[i],source_id = ['OSM'], lat = lines[i][1], 
#                              long = lines[i][0], country_code = countrycode, length=lengths[i],tags = tagdicts[i])
#        Ret_Data.PipeLines2.append(PipeLine2)
#

#        
#        
#        



"""Read funktion: OSM Data (pipelines) into K_Net Class"""

#######################################################################
#All functions are called from the makefile                           #
#######################################################################

from .M_PlotObjects import routelength

# from Code.OSM.M_OSM_CreateElements import M_OSM_CreateElements
from . import OSM_Pipeline_CountryCode as OP
from . M_OSM_PipeSegmenting import new_node_ID_nodes
from copy import deepcopy
from . import M_OSM_PipeSegmenting as PS

from unidecode import unidecode

# import configparser
# from  configparser import  ExtendedInterpolation
# from  pathlib            import Path
# import os
# import time

import json

# import Code.C_colors              as CC
from . import K_Netze as K_Netze
from . import K_Component as KC


def PipeSegments2Nodes(Pipelines):
    lat_list = []
    long_list = []
    id_list = []
    for pipe in Pipelines:
        for entry in zip(pipe.lat, pipe.long, pipe.node_id):
            lat_list.append(entry[0])
            long_list.append(entry[1])
            id_list.append(entry[2])
    return [lat_list, long_list, id_list]


def FindNodes4Pipelines(pipelines, nodes,new_nodes):
    for pipe in pipelines:
        for lat, long in zip(pipe.lat, pipe.long):
            for node in nodes:
                if lat == node.lat and long == node.long:
                    if node.id!=[]:
                        pipe.node_id.append(node.id)
                    #schaue in neu generierten punkten nach
                    else:
                        for new_node in new_nodes:
                            if lat == new_node.lat and long == new_node.long:
                                if new_node.id!=[]:
                                    pipe.node_id.append(new_node.id)
    pass


def PipeLine_length_error(pipelines, pipesegments):
    for pipe in pipelines:
        error = 0
        for segment in pipesegments:
            segment_id = segment.id.rpartition("_Seg_")[0]
            if pipe.id == segment_id:
                error += float(segment.uncertainty["length_km"])
        pipe.uncertainty["length_km"] = (round(error * 1000) / 1000)
    pass


def OSMways2net(elements, data, min_length=0):

    """ Convert OSMways to plotable lines """

    refs = []
    lines = []
    ids = []
    lengths = []
    tags = []
    tagdicts = []
    node_ids = []
    if "Way" in elements.keys():
        node_id_list = []
        for entry in elements["Way"]:
            refs = elements["Way"][entry]["refs"]
            lonlat_array = []
            node_id_list = []
            for ID in refs:
                lonlat_array.append(data["Node"][str(ID)]["lonlat"])
                node_id_list.append("OSM-" + str(ID))
            node_ids.append(node_id_list)
            long = []
            lat = []
            line = []

            for coords in lonlat_array:
                long.append(coords[0])
                lat.append(coords[1])
            linelength = float(routelength(long, lat))
            tag = (
                unidecode(
                    json.dumps(
                        elements["Way"][entry]["tags"], ensure_ascii=False, indent=2
                    )
                )
                .replace("{", "")
                .replace("}", "")
                .replace('\\"', "")
            )
            tagdict = elements["Way"][entry]["tags"]
            line = [long, lat]
            # only lines longer than min_length
            if linelength >= min_length:
                lines.append(line)
                lengths.append(float(linelength))
                ids.append("OSM-" + entry)
                tags.append(tag)
                tagdicts.append(tagdict)
    #                node_ids.append(refs)
    return lines, ids, lengths, tags, node_ids, tagdicts


def OSMnodes2net(elements):

    """ Convert OSMways to plotable lines """

    nodes = []
    ids = []
    tags = []
    node_ids = []
    tagdicts = []

    if "Node" in elements.keys():
        for entry in elements["Node"]:
            idd = elements["Node"][entry]["id"]
            node_ids.append(idd)
            long = elements["Node"][entry]["lonlat"][0]
            lat = elements["Node"][entry]["lonlat"][1]
            tag = (
                unidecode(
                    json.dumps(
                        elements["Node"][entry]["tags"], ensure_ascii=False, indent=2
                    )
                )
                .replace("{", "")
                .replace("}", "")
                .replace('\\"', "")
            )
            tagdict = elements["Node"][entry]["tags"]
            tagdicts.append(tagdict)
            node = [long, lat]
            nodes.append(node)
            ids.append(entry)
            tags.append(tag)

    return nodes, ids, tags, node_ids, tagdicts


def OSMways2nodes(elements, data, min_length=0):

    """Convert OSM Pipeline endpoints to Nodes"""

    refs = set()
    if "Way" in elements:
        refs = set(
            ref for way_id, entry in elements["Way"].items() for ref in entry["refs"]
        )

    refs = list(refs)
    new_ids = set("OSM-%s" % ref for ref in refs)
    points = {}

    for point in data["Node"]:
        point_id = "OSM-%s" % data["Node"][point]["id"]
        if point_id in new_ids:
            data["Node"][point]["id"] = point_id
            points[point_id] = data["Node"][point]

    return {"Node": points}


def extract_lat_longs(Component):
    lat_longs = []

    for pipe in Component:
        for pipelat, pipelong in zip(pipe.lat, pipe.long):
            lat_longs.append((pipelat, pipelong))
    return set(lat_longs)


def RawData2ClassData(
    Data,
    elements,
    JSON_outputfile,
    TM_World_Borders_file,
    countrycode,
    CreateCountryCodeForLines=False,
    LoadCountryCodeForLines=False,
    min_length=1,
    segment_length=100,
    verbose=False,
):

    Ret_Data = K_Netze.NetComp_OSM()

    ########
    # Pipeline
    ########
    print("---Load Pipelines---")
    [lines, ids, lengths, tags, node_ids, tagdicts] = OSMways2net(
        elements["pipelines"], Data, min_length=min_length
    )
    # pipenodes_dict=OSMways2nodes(elements["pipelines"],Data,min_length=0)
    i = 0
    length_uncertainty_factor = 0.3
    if segment_length == 0:
        length_uncertainty_factor = 0
    for i, entry in enumerate(zip(lines, ids, lengths, tags)):
        PipeLine = KC.OSMComponent(
            id=ids[i],
            name=ids[i],
            node_id=[],  # node_ids[i],
            source_id=["OSM"],
            lat=lines[i][1],
            long=lines[i][0],
            country_code=countrycode,
            tags=tagdicts[i],
            param={"length_km": float(lengths[i]),"path_lat":lines[i][1],"path_long":lines[i][0]},
            method={"length_km": "SciGRID_gas"},
            uncertainty={},
        )
        # ,                                 length=lengths[i])

        Ret_Data.PipeLines.append(PipeLine)


    # here wieder rein
    deepcopy(Ret_Data.PipeLines)
    Ret_Data.PipeSegments = PS.OSM_PipelineSegmenting(
        deepcopy(Ret_Data.PipeLines), length=segment_length, minlength=min_length
    )
    Ret_Data.PipeLines = PS.OSM_PipelineSegmenting(
        Ret_Data.PipeLines, length=segment_length, minlength=min_length, composed=True
    )
    # for pipe in Ret_Data.PipeLines:
        # print(pipe.param)
        # print(pipe.lat)
    for pipe in Ret_Data.PipeLines:
        pipe.param["length_km"] = pipe.getPipeLength()
    node_coords = extract_lat_longs(Ret_Data.PipeSegments)
    new_ID_generator = new_node_ID_nodes()
    node_list_ids = [next(new_ID_generator) for node in node_coords]
    nodes_id_dict = {}
    for coord, node_id in zip(node_coords, node_list_ids):
        nodes_id_dict.update({coord: node_id})

    for pipe in Ret_Data.PipeSegments:
        first_node = (pipe.lat[0], pipe.long[0])
        second_node = (pipe.lat[1], pipe.long[1])
        node_id = [nodes_id_dict[first_node], nodes_id_dict[second_node]]
        pipe.node_id = node_id

    for pipe in Ret_Data.PipeSegments:
        length = pipe.param["length_km"]
        pipe.uncertainty["length_km"] =  length_uncertainty_factor * float(length)

    print("---Load Nodes---")
    pipenodes = OSMways2nodes(elements["pipelines"], Data, min_length=min_length)
    # print(len(pipenodes['Node']))
    # print(pipenodes['Node'])
    [nodes, ids, tags, node_ids, tagdicts] = OSMnodes2net(pipenodes)

    for i, entry in enumerate(zip(nodes, ids, tags, node_ids, tagdicts)):
        Node = KC.OSMComponent(
            id=ids[i],
            name=ids[i],
            node_id=[node_ids[i]],
            source_id=["OSM"],
            lat=nodes[i][1],
            long=nodes[i][0],
            country_code=countrycode,
            param={},
            uncertainty={},
            method={},
            tags=tagdicts[i],
        )

        Ret_Data.Nodes.append(Node)


    pipelist = PipeSegments2Nodes(Ret_Data.PipeSegments)

    # transpose
    pipelist = list(map(list, zip(*pipelist)))

    # unique
    pipelist = [tuple(row) for row in pipelist]
    pipelist = set(pipelist)
    # pipelist=[list(row) for row in pipelist]
    # transpose back
    pipelist = list(map(list, zip(*pipelist)))

    lats, longs, ids = pipelist

    for i, entry in enumerate(zip(lats, longs, ids)):
        PipePoint = KC.OSMComponent(
            id=ids[i],
            name=ids[i],
            node_id=[],
            source_id=["OSM_Seq"],
            lat=lats[i],
            long=longs[i],
            country_code=countrycode,
            param={},
            uncertainty={},
            method={},
            tags={},
        )
        Ret_Data.PipePoints.append(PipePoint)

    #PipePoints are only newly created points
    Ret_Data.PipePoints = sorted(
        Ret_Data.PipePoints, key=lambda PipePoints: int(PipePoints.id[4:])
    )
    
    FindNodes4Pipelines(Ret_Data.PipeLines, Ret_Data.Nodes,Ret_Data.PipePoints)
    #Merging new generated points to point table
    Ret_Data.Nodes=Ret_Data.Nodes+Ret_Data.PipePoints
    PipeLine_length_error(Ret_Data.PipeLines, Ret_Data.PipeSegments)
    if CreateCountryCodeForLines == True:

        OP.Create_Pipelines_CountryCodes(
            JSON_outputfile, TM_World_Borders_file, Ret_Data, 
            countrycode, verbose=False)

    if LoadCountryCodeForLines == True:
        print("---Load Line Countrycodes--")
        OP.Load_Pipelines_Countrycodes(JSON_outputfile, Ret_Data, 
                                       countrycode)
        OP.Load_Pipesegments_Countrycodes(JSON_outputfile, Ret_Data, 
                                                countrycode)
                  
    return Ret_Data






import numpy as np
import pandas as pd
import os

def CSV_Write(file, subkeys, subvalues):
    """Description:
    -------------
        Schreibt die Eigenschaften(subkeys) eines Netzattribute in ein File
        zusammen mit seinen werten (subvalues)
        file = filepath+name.csv (des Netzattributes)
    Input:
    ------
        file        String containing absolute path and name to file
        subkeys
        subvalues
    Called by:
    ----------
        WriteCSVfiles()

    Needs:
    ------
        Panda, Numpy
    """

    tsubvalues  = np.matrix(subvalues,dtype=object)
    tsubvalues  = np.transpose(tsubvalues)
    df          = pd.DataFrame(tsubvalues, subkeys,dtype=object)
    df.T.to_csv(file,sep=';', index = False, na_rep = 'None')


def write(Grid,NameStart,filepath):
    """Writes Instance of Netz-Class (**Grid**) into CSV tables, located in relative folder **RelDirName**.

..comments:
    Input:
        CSV_Path_write:  Output Directory
        Grid:          Instance of Netz Class
    Return:
        []
    """
    # dir name
    RelDirName=filepath
    dataFolder = os.getcwd()
    DirName = os.path.join(dataFolder, RelDirName)

    FileList = Grid.CompLabels()
    
    if not os.path.exists(os.path.dirname(RelDirName)):
        os.makedirs(os.path.dirname(RelDirName))

    
#    list(map(os.unlink, (os.path.join( DirName,f) for f in os.listdir(DirName)) ) )

    for key in FileList:
        # Resetting values
        subkeys             = []
        dictSubKeysParam    = []
        dictSubKeysMethod   = []
        subvalues           = []
        tempsubvalues       = []
        i                   = 0
        dictSubKeysUncertainty = []
        # groing through each component, as long as length larger than 0
        if len(Grid.__dict__[key]) > 0:
            for subkey in Grid.__dict__[key][0].__dict__.keys():
                if 'param' in subkey:
                    for subKey2 in Grid.__dict__[key][0].param.keys():
                        subkeys.append(subKey2)
                        dictSubKeysParam.append(subKey2)
                elif 'uncertainty' in subkey:
                    for subKey2 in Grid.__dict__[key][0].uncertainty.keys():
                        subkeys.append(subKey2)
                        dictSubKeysUncertainty.append(subKey2)
                elif 'method' in subkey:
                    for subKey2 in Grid.__dict__[key][0].method.keys():
                        subkeys.append(subKey2)
                        dictSubKeysMethod.append(subKey2)
                elif 'tags' in subkey:
                    for subKey2 in Grid.__dict__[key][0].tags.keys():
                        subkeys.append(subKey2)
                        dictSubKeysMethod.append(subKey2)
                elif 'license' in subkey:
                    for subKey2 in Grid.__dict__[key][0].tags.keys():
                        subkeys.append(subKey2)
                        dictSubKeysMethod.append(subKey2)
                else:
                    subkeys.append(subkey)

            # Loop for number of Components
            fileSubKeys = []
            if len(Grid.__dict__[key]) > 0:
                subkeys         = Grid.__dict__[key][i].__dict__.keys()
                for ss in subkeys:
                    fileSubKeys.append(ss)

            for i in range(len(Grid.__dict__[key])):
                subvalue        = Grid.__dict__[key][i].__dict__.values()
                tempsubvalues   = []
                for val, subkey in zip(subvalue, subkeys):
                    tempsubvalues.append(val)

                subvalues.append(tempsubvalues)
            if 'Meta_' in key:
                # Writing the compopnent meta data
                CSV_Write(os.path.join((DirName), key + '.csv'), fileSubKeys, subvalues)
            else:
                # Writing the component position data
                CSV_Write(os.path.join((DirName), NameStart + '_' + key + '.csv'), fileSubKeys, subvalues)
        else:
            pass

